from ...discord.client import Client
from ...discord.events.core import DEFAULT_EVENT_HANDLER
from ...discord.events.handling_helpers import EventWaitforBase, asynclist

from .bases import *
from .choose_menu import *
from .closer import *
from .io import *
from .pagination import *
from .user_menu import *
from .utils import *
from .waiters import *

__all__ = (
    'setup_ext_command_utils',
    *bases.__all__,
    *choose_menu.__all__,
    *closer.__all__,
    *io.__all__,
    *pagination.__all__,
    *utils.__all__,
    *waiters.__all__,
    *user_menu.__all__,
)

from .. import register_library_extension, register_setup_function

def setup_ext_command_utils(client):
    """
    Setups the `command_utils` extension on the given client by adding ``ReactionAddWaitfor``, ``ReactionDeleteWaitfor``
    and ``MessageCreateWaitfor`` to it.
    
    Parameters
    ----------
    client : ``Client``
        The client to setup the extension on.
    """
    if not isinstance(client, Client):
        raise TypeError(
            f'Expected type `{Client.__name__}` as client, got {client.__class__.__name__}; {client!r}.'
        )
    
    event_reaction_add = client.events.reaction_add
    while True:
        if event_reaction_add is DEFAULT_EVENT_HANDLER:
            client.events(ReactionAddWaitfor)
            break
        
        if type(event_reaction_add) is asynclist:
            for event in list.__iter__(event_reaction_add):
                if isinstance(event, EventWaitforBase):
                    break
            else:
                client.events(ReactionAddWaitfor)
            
            break
        
        if isinstance(event_reaction_add, EventWaitforBase):
            break
        
        client.events(ReactionAddWaitfor)
        break
    
    event_reaction_delete = client.events.reaction_delete
    while True:
        if event_reaction_delete is DEFAULT_EVENT_HANDLER:
            client.events(ReactionDeleteWaitfor)
            break
        
        if type(event_reaction_delete) is asynclist:
            for event in list.__iter__(event_reaction_add):
                if isinstance(event, EventWaitforBase):
                    break
            else:
                client.events(ReactionDeleteWaitfor)
            
            break
        
        if isinstance(event_reaction_delete, EventWaitforBase):
            break
        
        client.events(ReactionDeleteWaitfor)
        break
    
    event_message_create = client.events.message_create
    while True:
        if event_message_create is DEFAULT_EVENT_HANDLER:
            client.events(MessageCreateWaitfor)
            break
        
        if type(event_message_create) is asynclist:
            for event in list.__iter__(event_message_create):
                if isinstance(event, EventWaitforBase):
                    break
            else:
                client.events(MessageCreateWaitfor)
            
            break
        
        if isinstance(event_message_create, EventWaitforBase):
            break
        
        client.events(MessageCreateWaitfor)
        break


register_library_extension('HuyaneMatsu.command_utils')

register_setup_function(
    'HuyaneMatsu.command_utils',
    setup_ext_command_utils,
    None,
    None,
)



__all__ = ('ChannelForum',)


from scarletio import copy_docs
from scarletio import export, include

from ..core import CHANNELS
from ..permission import Permission
from ..permission.permission import PERMISSION_MASK_VIEW_CHANNEL, PERMISSION_NONE, PERMISSION_THREAD_AND_VOICE_DENY
from ..preconverters import preconvert_int_options, preconvert_snowflake, preconvert_str

from . import channel_types as CHANNEL_TYPES
from .channel_base import ChannelBase
from .channel_guild_base import ChannelGuildMainBase
from .channel_thread import AUTO_ARCHIVE_DEFAULT, AUTO_ARCHIVE_OPTIONS


parse_permission_overwrites = include('parse_permission_overwrites')


@export
class ChannelForum(ChannelGuildMainBase):
    """
    Represents a ``Guild`` forum channel.
    
    Attributes
    ----------
    id : `int`
        Unique identifier of the channel.
    _permission_cache : `None`, `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    parent_id : `0`
        The channel's parent's identifier.
    guild_id : `int`
        The channel's guild's identifier.
    name : `str`
        The channel's name.
    permission_overwrites : `dict` of (`int`, ``PermissionOverwrite``) items
        The channel's permission overwrites.
    position : `int`
        The channel's position.
    default_auto_archive_after : `int`
        The default duration (in seconds) for newly created threads to automatically archive the themselves. Defaults
        to `3600`. Can be one of: `3600`, `86400`, `259200`, `604800`.
    
    Class Attributes
    ----------------
    DEFAULT_TYPE : `int` = `15`
        The preferred channel type, if there is no channel type included.
    INTERCHANGE : `tuple` of `int` = `(15,)`
        Defines to which channel type this channel's type can be interchanged. The channel's direct type must be of
        them.
    ORDER_GROUP : `int` = `0`
        An order group what defined which guild channel type comes after the other one.
    type : `int` = `15`
        The channel's Discord side type.
    """
    __slots__ = ('default_auto_archive_after', )
    
    ORDER_GROUP = 0
    DEFAULT_TYPE = CHANNEL_TYPES.guild_forum
    INTERCHANGE = (CHANNEL_TYPES.guild_forum, )
    type = CHANNEL_TYPES.guild_forum
    
    
    def __new__(cls, data, client, guild_id):
        """
        Creates a forum channel from the channel data received from Discord. If the channel already exists and if
        it is partial, then updates it.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data received from Discord.
        client : `None`, ``Client``
            The client, who received the channel's data, if any.
        guild_id : `None`, ``Guild``
            The channel's guild's identifier.
        """
        channel_id = int(data['id'])
        try:
            self = CHANNELS[channel_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = channel_id
            CHANNELS[channel_id] = self
        else:
            if self.clients:
                return self
        
        self._permission_cache = None
        self.name = data['name']
        
        self._init_parent_and_position(data, guild_id)
        self.permission_overwrites = parse_permission_overwrites(data)
        
        
        default_auto_archive_after = data.get('default_auto_archive_duration', None)
        if default_auto_archive_after is None:
            default_auto_archive_after = AUTO_ARCHIVE_DEFAULT
        else:
            default_auto_archive_after *= 60
        self.default_auto_archive_after = default_auto_archive_after
        
        return self
    
    
    @classmethod
    @copy_docs(ChannelBase._create_empty)
    def _create_empty(cls, channel_id, channel_type, guild_id):
        self = super(ChannelForum, cls)._create_empty(channel_id, channel_type, guild_id)
        
        self.default_auto_archive_after = AUTO_ARCHIVE_DEFAULT
        
        return self
    
    
    @property
    @copy_docs(ChannelBase.display_name)
    def display_name(self):
        return self.name.lower()
    
    
    @copy_docs(ChannelBase._update_attributes)
    def _update_attributes(self, data):
        self._permission_cache = None
        self._set_parent_and_position(data)
        self.permission_overwrites = parse_permission_overwrites(data)
        
        self.name = data['name']
        
        default_auto_archive_after = data.get('default_auto_archive_duration', None)
        if default_auto_archive_after is None:
            default_auto_archive_after = AUTO_ARCHIVE_DEFAULT
        else:
            default_auto_archive_after *= 60
        self.default_auto_archive_after = default_auto_archive_after
    
    
    def _difference_update_attributes(self,data):
        """
        Updates the channel and returns it's overwritten attributes as a `dict` with a `attribute-name` - `old-value`
        relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data received from Discord.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            All item in the returned dict is optional.
        
        Returned Data Structure
        -----------------------
        
        +-------------------------------+---------------------------------------------------+
        | Keys                          | Values                                            |
        +===============================+===================================================+
        | default_auto_archive_after    | `int`                                             |
        +-------------------------------+---------------------------------------------------+
        | parent_id                     | `int`                                             |
        +-------------------------------+---------------------------------------------------+
        | name                          | `str`                                             |
        +-------------------------------+---------------------------------------------------+
        | permission_overwrites         | `dict` of (`int`, ``PermissionOverwrite``) items  |
        +-------------------------------+---------------------------------------------------+
        | position                      | `int`                                             |
        +-------------------------------+---------------------------------------------------+
        """
        self._permission_cache = None
        old_attributes = {}
        
        name = data['name']
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        permission_overwrites = parse_permission_overwrites(data)
        if self.permission_overwrites != permission_overwrites:
            old_attributes['permission_overwrites'] = self.permission_overwrites
            self.permission_overwrites = permission_overwrites
        
        self._update_parent_and_position(data, old_attributes)
        
        default_auto_archive_after = data.get('default_auto_archive_duration', None)
        if default_auto_archive_after is None:
            default_auto_archive_after = AUTO_ARCHIVE_DEFAULT
        else:
            default_auto_archive_after *= 60
        if self.default_auto_archive_after != default_auto_archive_after:
            old_attributes['default_auto_archive_after'] = self.default_auto_archive_after
            self.default_auto_archive_after = default_auto_archive_after
        
        return old_attributes
    
    
    @copy_docs(ChannelBase.permissions_for)
    def permissions_for(self, user):
        result = self._permissions_for(user)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # forum channels do not have thread and voice related permissions
        result &= PERMISSION_THREAD_AND_VOICE_DENY
        
        return Permission(result)
    
    
    @copy_docs(ChannelBase.permissions_for_roles)
    def permissions_for_roles(self, *roles):
        result = self._permissions_for_roles(roles)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # forum channels do not have thread and voice related permissions
        result &= PERMISSION_THREAD_AND_VOICE_DENY
        return Permission(result)
    
    @classmethod
    def precreate(cls, channel_id, **kwargs):
        """
        Precreates the channel by creating a partial one with the given parameters. When the channel is loaded
        the precreated channel will be picked up. If an already existing channel would be precreated, returns that
        instead and updates that only, if that is a partial channel.
        
        Parameters
        ----------
        channel_id : `int`, `str`
            The channel's id.
        **kwargs : keyword parameters
            Additional predefined attributes for the channel.
        
        Other Parameters
        ----------------
        default_auto_archive_after : `int`, Optional (Keyword only)
            The channel's ``.default_auto_archive_after``.
        name : `str`, Optional (Keyword only)
            The channel's ``.name``.
        
        Returns
        -------
        channel : ``ChannelForum``
        
        Raises
        ------
        TypeError
            If any parameter's type is bad or if unexpected parameter is passed.
        ValueError
            If an parameter's type is good, but it's value is unacceptable.
        """
        channel_id = preconvert_snowflake(channel_id, 'channel_id')
        
        if kwargs:
            processable = []
            
            try:
                name = kwargs.pop('name')
            except KeyError:
                pass
            else:
                name = preconvert_str(name, 'name', 2, 100)
                processable.append(('name', name))
            
            try:
                default_auto_archive_after = kwargs.pop('default_auto_archive_duration')
            except KeyError:
                pass
            else:
                default_auto_archive_after = preconvert_int_options(default_auto_archive_after,
                    'default_auto_archive_after', AUTO_ARCHIVE_OPTIONS)
                processable.append(('default_auto_archive_after', default_auto_archive_after))
            
            if kwargs:
                raise TypeError(f'Unused or unsettable attributes: {kwargs!r}.')
        
        else:
            processable = None
        
        try:
            self = CHANNELS[channel_id]
        except KeyError:
            self = cls._create_empty(channel_id, cls.type, 0)
            CHANNELS[channel_id] = self
            
        else:
            if not self.partial:
                return self
        
        if (processable is not None):
            for item in processable:
                setattr(self, *item)
        
        return self


    @copy_docs(ChannelGuildMainBase.to_data)
    def to_data(self):
        # same as `ChannelGuildMainBase`'s. Keep this function here for reference.
        data = ChannelGuildMainBase.to_data(self)
        
        # default_auto_archive_duration
        data['default_auto_archive_duration'] = self.default_auto_archive_after // 60
        
        return data



__all__ = ()

from scarletio import KeepType

from ...discord.client.utils import ClientWrapper
from ...discord.events.handling_helpers import _EventHandlerManagerRouter

from .application_command import SlasherApplicationCommand
from .slasher import Slasher


def interactions_getter(manager_router):
    """
    Gets the slash command processor using `Client.slasher` of an ``_EventHandlerManagerRouter``.
    
    Parameters
    ----------
    manager_router : ``_EventHandlerManagerRouter``
        The caller manager router.
    
    Returns
    -------
    handlers : `list` of ``Slasher``
    """
    handlers = []
    for client in manager_router.parent.clients:
        manager = getattr(client, 'interactions', None)
        if manager is None:
            continue
        
        handler = manager.parent
        if isinstance(handler, Slasher):
            handlers.append(handler)
    
    return handlers


def from_class_constructor(klass):
    """
    Creates a slash command from the given class.
    
    Raises
    ------
    BaseException
        Any exception raised by the respective ``SlasherApplicationCommand`` constructor.
    """
    return SlasherApplicationCommand.from_class(klass)


@KeepType(ClientWrapper)
class ClientWrapper:
    
    @property
    def interactions(self):
        """
        Returns a ``_EventHandlerManagerRouter``, with what slash commands can be added to more clients at the
        same time.
        
        Returns
        -------
        event_handler_manager_router : ``_EventHandlerManagerRouter``
        """
        return _EventHandlerManagerRouter(self, interactions_getter, from_class_constructor)




__all__ = ('ScheduledEventSubscribeEvent', 'ScheduledEventUnsubscribeEvent')

from scarletio import copy_docs

from ..bases import EventBase


class ScheduledEventSubscribeEvent(EventBase):
    """
    Represents a `GUILD_SCHEDULED_EVENT_USER_ADD` event.
    
    Attributes
    ----------
    guild_id : `int`
        The guild's identifier, where the event will be.
    scheduled_event_id : `int`
        The scheduled event's identifier.
    user_id : `int`
        The identifier of the user, who subscribed to the event.
    """
    __slots__ = ('guild_id', 'scheduled_event_id', 'user_id', )
    
    def __new__(cls, data):
        """
        Creates a new scheduled event (un)subscribe event from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Scheduled event subscribe event data.
        """
        guild_id = int(data['guild_id'])
        scheduled_event_id = int(data['guild_scheduled_event_id'])
        user_id = int(data['user_id'])
        
        self = object.__new__(cls)
        self.guild_id = guild_id
        self.scheduled_event_id = scheduled_event_id
        self.user_id = user_id
        return self
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' guild_id=',
            repr(self.guild_id),
            ', scheduled_event_id=',
            repr(self.scheduled_event_id),
            ', user_id=',
            repr(self.scheduled_event_id),
            '>'
        ]
        
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 3
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.guild_id
        yield self.scheduled_event_id
        yield self.user_id
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.guild_id != other.guild_id:
            return False
        
        if self.scheduled_event_id != other.scheduled_event_id:
            return False
        
        if self.user_id != other.user_id:
            return False
        
        return True
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        return self.guild_id ^ self.scheduled_event_id ^ self.user_id


class ScheduledEventUnsubscribeEvent(ScheduledEventSubscribeEvent):
    """
    Represents a `GUILD_SCHEDULED_EVENT_USER_REMOVE` event.
    
    Attributes
    ----------
    guild_id : `int`
        The guild's identifier, where the event will be.
    scheduled_event_id : `int`
        The scheduled event's identifier.
    user_id : `int`
        The identifier of the user, who unsubscribed to the event.
    """
    __slots__ = ()



import warnings

warnings.warn(
    '`hata.ext.prettyprint` is discontinued and will be removed in 2020 Jun.',
    FutureWarning,
)

from .prettyprint import *

__all__ = prettyprint.__all__

from .. import register_library_extension
register_library_extension('HuyaneMatsu.prettyprint')
del register_library_extension



__all__ = ('TopGGClient', )

from scarletio import Future, Task, shield, sleep
from scarletio import IgnoreCaseMultiValueDictionary, LOOP_TIME, WeakReferer, from_json, to_json
from scarletio.http_client import RequestContextManager
from scarletio.web_common.headers import AUTHORIZATION, CONTENT_TYPE, METHOD_GET, METHOD_POST, RETRY_AFTER, USER_AGENT

from ...discord.client import Client
from ...discord.core import KOKORO
from ...discord.http import LIBRARY_USER_AGENT

from .bots_query import create_bots_query_search_value, get_bots_query_sort_by_value
from .constants import (
    JSON_KEY_POST_BOT_STATS_GUILD_COUNT, JSON_KEY_POST_BOT_STATS_SHARD_COUNT, JSON_KEY_POST_BOT_STATS_SHARD_ID,
    JSON_KEY_VOTED, JSON_KEY_WEEKEND_STATUS, QUERY_KEY_GET_BOTS_LIMIT, QUERY_KEY_GET_BOTS_OFFSET,
    QUERY_KEY_GET_BOTS_SEARCH_QUERY, QUERY_KEY_GET_BOTS_SORT_BY, QUERY_KEY_GET_USER_VOTE_USER_ID,
    RATE_LIMIT_BOTS_RESET_AFTER, RATE_LIMIT_BOTS_SIZE, RATE_LIMIT_GLOBAL_DEFAULT_DURATION,
    RATE_LIMIT_GLOBAL_RESET_AFTER, RATE_LIMIT_GLOBAL_SIZE
)
from .exceptions import TopGGGloballyRateLimited, TopGGHttpException
from .rate_limit_handling import RateLimitGroup, RateLimitHandler, StackedRateLimitHandler
from .types import BotInfo, BotsQueryResult, UserInfo


AUTO_POST_INTERVAL = 1800.0
WEEKEND_STATE_UPDATE_INTERVAL = 300.0

TOP_GG_ENDPOINT = 'https://top.gg/api'

async def _start_auto_post(client):
    """
    Client launch event handler starting auto post.
    
    This method is a coroutine.
    
    Parameters
    ----------
    client : ``Client``
    """
    top_gg_client = client.top_gg
    
    # update client is if required
    top_gg_client.client_id = client.id
    
    # Do not post initially, we might not have all the guilds loaded yet.
    if top_gg_client._auto_post_running:
        top_gg_client._auto_post_handler = KOKORO.call_later(
            AUTO_POST_INTERVAL,
            _trigger_auto_post,
            WeakReferer(top_gg_client),
        )


async def _stop_auto_post(client):
    """
    Client disconnect event handler, which stops auto posting.
    
    This method is a coroutine.
    
    Parameters
    ----------
    client : ``Client``
    """
    top_gg_client = client.top_gg
    
    top_gg_client._auto_post_running = False
    
    auto_post_handler = top_gg_client._auto_post_handler
    if (auto_post_handler is not None):
        top_gg_client._auto_post_handler = None
        auto_post_handler.cancel()


def _trigger_auto_post(top_gg_client_reference):
    """
    Triggers an auto post.
    
    Parameters
    ----------
    top_gg_client_reference : ``WeakReferer`` to ``TopGGClient``
        Weak reference to the top.gg client.
    """
    top_gg_client = top_gg_client_reference()
    if (top_gg_client is not None):
        Task(_do_auto_post(top_gg_client, top_gg_client_reference), KOKORO)


async def _do_auto_post(top_gg_client, top_gg_client_reference):
    """
    Does an auto post.
    
    This method is a coroutine.
    
    Parameters
    ----------
    top_gg_client : ``TopGGClient``
        The top.gg client.
    top_gg_client_reference : ``WeakReferer`` to ``TopGGClient``
        Weak reference to the top.gg client.
    
    Raises
    ------
    TopGGHttpException
        Any exception raised by top.gg api.
    """
    try:
        await top_gg_client.post_bot_stats()
    except ConnectionError:
        return
    finally:
        if top_gg_client._auto_post_running:
            top_gg_client._auto_post_handler = KOKORO.call_later(
                AUTO_POST_INTERVAL,
                _trigger_auto_post,
                top_gg_client_reference,
            )


async def get_weekend_status_task(top_gg_client):
    """
    Returns the weekend multiplier is on.
    
    This function is used for synchronizing multiple ``TopGGClient.get_weekend_status`` calls.
    
    This function is a coroutine.
    
    Parameters
    ----------
    top_gg_client : ``TopGGClient``
        The top.gg client to use.
    
    Returns
    -------
    weekend_status : `bool`
    
    Raises
    ------
    ConnectionError
        No internet connection.
    TopGGGloballyRateLimited
        If the client got globally rate limited by top.gg and `raise_on_top_gg_global_rate_limit` was given as
        `True`.
    TopGGHttpException
        Any exception raised by top.gg api.
    """
    try:
        data = await top_gg_client._get_weekend_status()
    finally:
        top_gg_client._weekend_status_request_task = None
    
    return data[JSON_KEY_WEEKEND_STATUS]


class TopGGClient:
    """
    Represents a client connection towards top.gg.
    
    Attributes
    ----------
    _auto_post_handler : `None`, ``TimerHandle``
        Handle to repeat the auto poster.
    _auto_post_running : `bool`
        Whether auto posting is still running.
    _global_rate_limit_expires_at : `float`
        When the global rate limit expires in monotonic time.
    _headers : `IgnoreCaseMultiValueDictionary`
        Request headers.
    _raise_on_top_gg_global_rate_limit : `bool`
        Whether ``TopGGGloballyRateLimited`` should be raised when the client gets globally rate limited.
    _rate_limit_handler_global : ``RateLimitHandler``
        Rate limit handler applied to all rate limits.
    _rate_limit_handler_bots : ``RateLimitHandler``
        Rate limit handler applied to `/bots` endpoints.
    _weekend_status_cache_time : `float`
        When the last ``.get_weekend_status`` was done.
    _weekend_status_cache_value : `bool`
        The response of the last ``get_weekend_status`` call.
    _weekend_status_request_task : `None`, ``Task``
        Synchronization task for requesting weekend status.
    client_id : `int`
        The client's identifier.
    client_reference : ``WeakReferer`` to ``Client``
        Weakreference towards the wrapped bot.
    http : ``DiscordHTTPClient``
        Http client to do requests with.
    top_gg_token : `str`
        Top.gg api token.
    """
    __slots__ = (
        '__weakref__', '_auto_post_handler', '_auto_post_running', '_global_rate_limit_expires_at', '_headers',
        '_raise_on_top_gg_global_rate_limit', '_rate_limit_handler_bots', '_rate_limit_handler_global',
        '_weekend_status_cache_time', '_weekend_status_cache_value', '_weekend_status_request_task', 'client_id',
        'client_reference', 'http', 'top_gg_token'
    )
    
    def __new__(cls, client, top_gg_token, auto_post_bot_stats=True, raise_on_top_gg_global_rate_limit=False):
        """
        Creates a new top.gg client instance.
        
        Parameters
        ----------
        client : ``Client``
            The discord client.
        top_gg_token : `str`
            Top.gg api token.
        auto_post_bot_stats : `bool`, Optional
            Whether auto post should be started as the client launches up.
            
            Defaults to `True`.
        raise_on_top_gg_global_rate_limit : `bool`, Optional
            Whether ``TopGGGloballyRateLimited`` should be raised when the client gets globally rate limited.
            
            Defaults to `False`.
        
        Raises
        ------
        TypeError
            - If `client` is not ``Client``.
            - If `top_gg_token` is not `str`.
            - If `auto_post_bot_stats` is not `bool`.
            - If `raise_on_top_gg_global_rate_limit` is not `bool` isinstance.
        """
        if not isinstance(client, Client):
            raise TypeError(
                f'`client` can be `{Client.__name__}`, got {client.__class__.__name__}; {client!r}.'
            )
        
        if not isinstance(top_gg_token, str):
            raise TypeError(
                f'`top_gg_token` can be `str`, got {top_gg_token.__class__.__name__}; {top_gg_token!r}.'
            )
        
        if not isinstance(auto_post_bot_stats, bool):
            raise TypeError(
                f'`auto_post_bot_stats` can be `bool`, got '
                f'{auto_post_bot_stats.__class__.__name__}; {auto_post_bot_stats!r}.'
            )
        
        if not isinstance(raise_on_top_gg_global_rate_limit, bool):
            raise TypeError(
                f'`raise_on_top_gg_global_rate_limit` can be `bool`, got '
                f'{raise_on_top_gg_global_rate_limit.__class__.__name__}; {raise_on_top_gg_global_rate_limit!r}.'
            )
        
        
        client_reference = WeakReferer(client)
        
        headers = IgnoreCaseMultiValueDictionary()
        headers[USER_AGENT] = LIBRARY_USER_AGENT
        headers[AUTHORIZATION] = top_gg_token
        
        self = object.__new__(cls)
        self.client_reference = client_reference
        self._auto_post_handler = None
        self._auto_post_running = auto_post_bot_stats
        self._raise_on_top_gg_global_rate_limit = raise_on_top_gg_global_rate_limit
        self._headers = headers
        
        self.client_id = client.id
        self.http = client.http
        self.top_gg_token = top_gg_token
        
        self._global_rate_limit_expires_at = 0.0
        self._rate_limit_handler_global = RateLimitGroup(RATE_LIMIT_GLOBAL_SIZE, RATE_LIMIT_GLOBAL_RESET_AFTER)
        self._rate_limit_handler_bots = RateLimitGroup(RATE_LIMIT_BOTS_SIZE, RATE_LIMIT_BOTS_RESET_AFTER)
        
        self._weekend_status_cache_time = 0.0
        self._weekend_status_cache_value = False
        self._weekend_status_request_task = None
        
        return self
    
    
    def is_auto_posting(self):
        """
        Returns whether the bot stats are auto posted to top.gg
        
        Returns
        -------
        is_auto_posting : `bool`
        """
        return self._auto_post_running
    
    
    def start_auto_posting(self):
        """
        Starts auto posting bot stats.
        """
        if not self._auto_post_running:
            self._auto_post_running = True
            
            # We can only start it, if our client is running.
            client = self.client_reference()
            if (client is not None) and client.running:
                
                # Check edge case
                auto_post_handler = self._auto_post_handler
                if (auto_post_handler is None):
                    
                    self._auto_post_handler = KOKORO.call_later(
                        AUTO_POST_INTERVAL,
                        _trigger_auto_post,
                        WeakReferer(self),
                    )
    
    def stop_auto_posting(self):
        """
        Stops auto posting bot starts.
        """
        if self._auto_post_running:
            self._auto_post_running = False
            
            auto_post_handler = self._auto_post_handler
            if (auto_post_handler is not None):
                self._auto_post_handler = None
                auto_post_handler.cancel()
    
    
    async def post_bot_stats(self):
        """
        Posts your guild & shard count.
        
        This method is a coroutine.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        TopGGGloballyRateLimited
            If the client got globally rate limited by top.gg and `raise_on_top_gg_global_rate_limit` was given as
            `True`.
        TopGGHttpException
            Any exception raised by top.gg api.
        """
        client = self.client_reference()
        if (client is not None):
            guild_count = len(client.guilds)
            
            shard_count = client.shard_count
            if shard_count < 1:
                shard_count = 1
            
            data = {
                JSON_KEY_POST_BOT_STATS_GUILD_COUNT: guild_count,
                JSON_KEY_POST_BOT_STATS_SHARD_COUNT: shard_count,
                JSON_KEY_POST_BOT_STATS_SHARD_ID: 0,
            }
            
            await self._post_bot_stats(data)
    
    
    async def get_weekend_status(self, *, force_update=False):
        """
        Returns the weekend multiplier is on.
        
        This method is a coroutine.
        
        Parameters
        ----------
        force_update : `bool`, Optional (Keyword only)
            Whether the weekend status should be forcefully updated instead of using the cached one.
        
        Returns
        -------
        weekend_status : `bool`
        
        Raises
        ------
        ConnectionError
            No internet connection.
        TopGGGloballyRateLimited
            If the client got globally rate limited by top.gg and `raise_on_top_gg_global_rate_limit` was given as
            `True`.
        TopGGHttpException
            Any exception raised by top.gg api.
        """
        if force_update or (self._weekend_status_cache_time < LOOP_TIME()):
            task = self._weekend_status_request_task
            if task is None:
                task = Task(get_weekend_status_task(self), KOKORO)
                self._weekend_status_request_task = task
            
            weekend_status = await shield(task, KOKORO)
        else:
            weekend_status = self._weekend_status_cache_value
        
        return weekend_status
    
    
    async def get_bot_voters(self):
        """
        Returns the last 1000 voters.
        
        This method is a coroutine.
        
        Returns
        -------
        voters : `list` of ``UserInfo``
        
        Raises
        ------
        ConnectionError
            No internet connection.
        TopGGGloballyRateLimited
            If the client got globally rate limited by top.gg and `raise_on_top_gg_global_rate_limit` was given as
            `True`.
        TopGGHttpException
            Any exception raised by top.gg api.
        """
        user_datas = await self._get_bot_voters()
        return [UserInfo.from_data(user_data) for user_data in user_datas]
    
    
    async def get_bot_info(self):
        """
        Returns your bot's information.
        
        This method is a coroutine.
        
        Returns
        -------
        bot_info : ``BotInfo``
        
        Raises
        ------
        ConnectionError
            No internet connection.
        TopGGGloballyRateLimited
            If the client got globally rate limited by top.gg and `raise_on_top_gg_global_rate_limit` was given as
            `True`.
        TopGGHttpException
            Any exception raised by top.gg api.
        """
        data = await self._get_bot_info()
        return BotInfo.from_data(data)
    
    
    async def get_bots(self, *, limit=50, offset=0, sort_by=None, search=None):
        """
        Gets information about multiple bots.
        
        This method is a coroutine.
        
        Parameters
        ----------
        limit : `int`, Optional (Keyword only)
            The amount of bots to query. Defaults to `50`.
        offset : `int`, Optional (Keyword only)
            Query offset. Defaults to `0`
        sort_by : `None`, `str`, Optional (Keyword only)
            Which field to sort by the bots. Defaults to `None`.
        search : `None`, `dict` of (`str`, `Any`) items, Optional (Keyword only)
            Fields an expected values to search for.
        
        Returns
        -------
        get_bots_result : ``BotsQueryResult``
        
        Raises
        ------
        LookupError
            - If `sort_by` refers to a not existent field.
            - If `search` contains a not existent field.
        ConnectionError
            No internet connection.
        TopGGGloballyRateLimited
            If the client got globally rate limited by top.gg and `raise_on_top_gg_global_rate_limit` was given as
            `True`.
        TopGGHttpException
            Any exception raised by top.gg api.
        
        Keys
        ----
        `sort_by` fields and `search` field's keys might be the following:
        
        - banner_url
        - certified_at
        - discriminator
        - donate_bot_guild_id
        - featured_guild_ids
        - github_url
        - id
        - invite_url
        - is_certified
        - long_description
        - name
        - owner_id
        - owner_ids
        - prefix
        - short_description
        - support_server_invite_url
        - tags
        - upvotes
        - upvotes_monthly
        - vanity_url
        - website_url
        """
        if limit > 500:
            limit = 500
        
        query_sort_by_value = get_bots_query_sort_by_value(sort_by)
        query_search_value = create_bots_query_search_value(search)
        
        query_parameters = {
            QUERY_KEY_GET_BOTS_LIMIT: limit,
            QUERY_KEY_GET_BOTS_OFFSET: offset,
            QUERY_KEY_GET_BOTS_SORT_BY: query_sort_by_value,
            QUERY_KEY_GET_BOTS_SEARCH_QUERY: query_search_value,
            # QUERY_KEY_GET_BOTS_FIELDS: BOTS_QUERY_FIELDS_VALUE, # Defaults to all fields, so we just skip it
        }
        
        data = await self._get_bots(query_parameters)
        return BotsQueryResult.from_data(data)
    
    
    async def get_user_info(self, user_id):
        """
        Gets user info for the given user identifier.
        
        This method is a coroutine.
        
        Parameters
        ----------
        user_id : `int`
            The user's identifier to get.
        
        Returns
        -------
        user_info : ``UserInfo``
        
        Raises
        ------
        ConnectionError
            No internet connection.
        TopGGGloballyRateLimited
            If the client got globally rate limited by top.gg and `raise_on_top_gg_global_rate_limit` was given as
            `True`.
        TopGGHttpException
            Any exception raised by top.gg api.
        """
        data = await self._get_user_info(user_id)
        return UserInfo.from_data(data)
    
    
    async def get_user_vote(self, user_id):
        """
        Returns whether the user voted in the last 12 hours.
        
        This method is a coroutine.
        
        Parameters
        ----------
        user_id : `int`
            The user's identifier.
        
        Returns
        -------
        voted : `bool`
        
        Raises
        ------
        ConnectionError
            No internet connection.
        TopGGGloballyRateLimited
            If the client got globally rate limited by top.gg and `raise_on_top_gg_global_rate_limit` was given as
            `True`.
        TopGGHttpException
            Any exception raised by top.gg api.
        """
        data = await self._get_user_vote({QUERY_KEY_GET_USER_VOTE_USER_ID: user_id})
        return bool(data[JSON_KEY_VOTED])
    
    
    async def _post_bot_stats(self, data):
        """
        Posts bot stats to top.gg.
        
        This method is a coroutine.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Bot stats.
        
        Returns
        -------
        response_data : `Any`
        
        Raises
        ------
        ConnectionError
            No internet connection.
        TopGGGloballyRateLimited
            If the client got globally rate limited by top.gg and `raise_on_top_gg_global_rate_limit` was given as
            `True`.
        TopGGHttpException
            Any exception raised by top.gg api.
        """
        return await self._request(
            METHOD_POST,
            f'{TOP_GG_ENDPOINT}/bots/stats',
            StackedRateLimitHandler(self._rate_limit_handler_bots, self._rate_limit_handler_global),
            data = data,
        )
    
    
    async def _get_weekend_status(self):
        """
        Gets weekend status from top.gg.
        
        This method is a coroutine.
        
        Returns
        -------
        response_data : `Any`
        
        Raises
        ------
        ConnectionError
            No internet connection.
        TopGGGloballyRateLimited
            If the client got globally rate limited by top.gg and `raise_on_top_gg_global_rate_limit` was given as
            `True`.
        TopGGHttpException
            Any exception raised by top.gg api.
        """
        return await self._request(
            METHOD_GET,
            f'{TOP_GG_ENDPOINT}/weekend',
            RateLimitHandler(self._rate_limit_handler_global),
        )
    
    
    async def _get_bot_voters(self):
        """
        Gets the last 1000 voters.
        
        This method is a coroutine.
        
        Returns
        -------
        response_data : `Any`
        
        Raises
        ------
        ConnectionError
            No internet connection.
        TopGGGloballyRateLimited
            If the client got globally rate limited by top.gg and `raise_on_top_gg_global_rate_limit` was given as
            `True`.
        TopGGHttpException
            Any exception raised by top.gg api.
        """
        return await self._request(
            METHOD_GET,
            f'{TOP_GG_ENDPOINT}/bots/{self.client_id}/votes',
            StackedRateLimitHandler(self._rate_limit_handler_bots, self._rate_limit_handler_global),
        )
    
    
    async def _get_bot_info(self):
        """
        Gets bot information and returns it.
        
        This method is a coroutine.
        
        Returns
        -------
        response_data : `Any`
        
        Raises
        ------
        ConnectionError
            No internet connection.
        TopGGGloballyRateLimited
            If the client got globally rate limited by top.gg and `raise_on_top_gg_global_rate_limit` was given as
            `True`.
        TopGGHttpException
            Any exception raised by top.gg api.
        """
        return await self._request(
            METHOD_GET,
            f'{TOP_GG_ENDPOINT}/bots/{self.client_id}',
            StackedRateLimitHandler(self._rate_limit_handler_bots, self._rate_limit_handler_global),
        )
    
    
    async def _get_bots(self, query_parameters):
        """
        Gets information about multiple bots.
        
        This method is a coroutine.
        
        Parameters
        ----------
        query_parameters : `dict` of (`str`, `Any`) items
            Query parameters.
            
        Returns
        -------
        response_data : `Any`
        
        Raises
        ------
        ConnectionError
            No internet connection.
        TopGGGloballyRateLimited
            If the client got globally rate limited by top.gg and `raise_on_top_gg_global_rate_limit` was given as
            `True`.
        TopGGHttpException
            Any exception raised by top.gg api.
        """
        return await self._request(
            METHOD_GET,
            f'{TOP_GG_ENDPOINT}/bots',
            StackedRateLimitHandler(self._rate_limit_handler_bots, self._rate_limit_handler_global),
            query_parameters = query_parameters,
        )
    
    
    async def _get_user_info(self, user_id):
        """
        Gets user info for the given user identifier.
        
        This method is a coroutine.
        
        Parameters
        ----------
        user_id : `int`
            The user's identifier to get.
            
        Returns
        -------
        response_data : `Any`
        
        Raises
        ------
        ConnectionError
            No internet connection.
        TopGGGloballyRateLimited
            If the client got globally rate limited by top.gg and `raise_on_top_gg_global_rate_limit` was given as
            `True`.
        TopGGHttpException
            Any exception raised by top.gg api.
        """
        return await self._request(
            METHOD_GET,
            f'{TOP_GG_ENDPOINT}/users/{user_id}',
            RateLimitHandler(self._rate_limit_handler_global),
        )

    async def _get_user_vote(self, query_parameters):
        """
        Returns whether the user voted in the last 12 hours.
        
        This method is a coroutine.
        
        Parameters
        ----------
        query_parameters : `dict` of (`str`, `Any`) items
            Query parameters.
        
        Returns
        -------
        response_data : `Any`
        
        Raises
        ------
        ConnectionError
            No internet connection.
        TopGGGloballyRateLimited
            If the client got globally rate limited by top.gg and `raise_on_top_gg_global_rate_limit` was given as
            `True`.
        TopGGHttpException
            Any exception raised by top.gg api.
        """
        return await self._request(
            METHOD_GET,
            f'{TOP_GG_ENDPOINT}/bots/{self.client_id}/check',
            RateLimitHandler(self._rate_limit_handler_global),
            query_parameters = query_parameters,
        )
    
    async def _request(self, method, url, rate_limit_handler, data=None, query_parameters=None):
        """
        Does a request towards top.gg API.
        
        This method is a coroutine.
        
        Parameters
        ----------
        method : `str`
            Http method.
        url : `str`
            Endpoint to do request towards.
        data : `None`, `Any`, Optional
            Json serializable data.
        rate_limit_handler : ``RateLimitHandlerBase`
            Rate limit handle to handle rate limit as.
        query_parameters : `None`, `Any`, Optional
            Query parameters.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        TopGGGloballyRateLimited
            If the client got globally rate limited by top.gg and `raise_on_top_gg_global_rate_limit` was given as
            `True`.
        TopGGHttpException
            Any exception raised by top.gg api.
        """
        headers = self._headers.copy()
        
        if (data is not None):
            headers[CONTENT_TYPE] = 'application/json'
            data = to_json(data)
        
        try_again = 2
        while try_again > 0:
            global_rate_limit_expires_at = self._global_rate_limit_expires_at
            if global_rate_limit_expires_at > LOOP_TIME():
                if self._raise_on_top_gg_global_rate_limit:
                    raise TopGGGloballyRateLimited(None)
                
                future = Future(KOKORO)
                KOKORO.call_at(global_rate_limit_expires_at, Future.set_result_if_pending, future, None)
                await future
            
            async with rate_limit_handler.ctx():
                try:
                    async with RequestContextManager(
                        self.http._request(method, url, headers, data, query_parameters)
                    ) as response:
                        response_data = await response.text(encoding='utf-8')
                except OSError as err:
                    if not try_again:
                        raise ConnectionError('Invalid address or no connection with Top.gg.') from err
                    
                    await sleep(0.5 / try_again, KOKORO)
                    
                    try_again -= 1
                    continue
                
                response_headers = response.headers
                status = response.status
                
                content_type_headers = response_headers.get(CONTENT_TYPE, None)
                if (content_type_headers is not None) and content_type_headers.startswith('application/json'):
                    response_data = from_json(response_data)
                
                if 199 < status < 305:
                    return response_data
                
                # Are we rate limited?
                if status == 429:
                    try:
                        retry_after = headers[RETRY_AFTER]
                    except KeyError:
                        retry_after = RATE_LIMIT_GLOBAL_DEFAULT_DURATION
                    else:
                        try:
                            retry_after = float(retry_after)
                        except ValueError:
                            retry_after = RATE_LIMIT_GLOBAL_DEFAULT_DURATION
                    
                    self._global_rate_limit_expires_at = LOOP_TIME() + retry_after
                    
                    if self._raise_on_top_gg_global_rate_limit:
                        raise TopGGGloballyRateLimited(None)
                        
                    await sleep(retry_after, KOKORO)
                    continue
                
                
                # Python casts sets to frozensets
                if (status in {400, 401, 402, 404}):
                    raise TopGGHttpException(response, response_data)
                
                if try_again and (status >= 500):
                    await sleep(10.0 / try_again, KOKORO)
                    try_again -= 1
                    continue
                
                raise TopGGHttpException(response, response_data)



import sys
import warnings
from types import ModuleType

import scarletio
from scarletio import http_client as scarletio_http_client
from scarletio import web_common as scarletio_web_common
from scarletio import websocket as scarletio_websocket


class deprecate(ModuleType):
    def __new__(cls):
        frame = sys._getframe().f_back
        globals_ = frame.f_globals
        
        module_name = globals_['__spec__'].name
        
        module = ModuleType.__new__(cls, module_name)
        
        
        for key, value in globals_.items():
            setattr(module, key, value)
        
        sys.modules[module_name] = module
        
    
    def __getattr__(self, attribute_name):
        for module_name, module_value in (
            ('scarletio', scarletio),
            ('scarletio.web_common', scarletio_web_common),
            ('scarletio.http_client', scarletio_http_client),
            ('scarletio.websocket', scarletio_websocket),
        ):
            try:
                value = getattr(module_value, attribute_name)
            except AttributeError:
                continue
            
            warnings.warn(
                (
                    f'Hata\'s backend has been moved out to a separate library.\n'
                    f'Please do `from {module_name} import {attribute_name}` instead.'
                ),
                FutureWarning,
            )
            
            return value
        
        raise AttributeError(attribute_name)



__all__ = (
    'ApplicationCommandAutocompleteInteraction', 'ApplicationCommandAutocompleteInteractionOption',
    'ApplicationCommandInteraction', 'ApplicationCommandInteractionOption', 'ComponentInteraction',
    'FormSubmitInteraction', 'FormSubmitInteractionOption', 'InteractionEvent', 'InteractionResponseContext',
    'InteractionType'
)


import reprlib, warnings

from scarletio import Future, export, future_or_timeout, shield

from ..bases import DiscordEntity, EventBase
from ..channel import create_partial_channel_from_data
from ..core import (
    APPLICATION_ID_TO_CLIENT, CHANNELS, GUILDS, INTERACTION_EVENT_MESSAGE_WAITERS, INTERACTION_EVENT_RESPONSE_WAITERS,
    KOKORO
)
from ..guild import create_partial_guild_from_id
from ..message import Attachment, Message
from ..oauth2.helpers import parse_guild_locale, parse_locale
from ..permission import Permission
from ..permission.permission import PERMISSION_PRIVATE
from ..role import Role
from ..user import ClientUserBase, User

from .components import ComponentBase
from .preinstanced import ApplicationCommandOptionType, ComponentType, InteractionType


RESPONSE_FLAG_DEFERRING = 1 << 0
RESPONSE_FLAG_DEFERRED = 1 << 1
RESPONSE_FLAG_RESPONDING = 1 << 2
RESPONSE_FLAG_RESPONDED = 1 << 3
RESPONSE_FLAG_EPHEMERAL = 1 << 4

RESPONSE_FLAG_NONE = 0
RESPONSE_FLAG_ACKNOWLEDGING = RESPONSE_FLAG_DEFERRING | RESPONSE_FLAG_RESPONDING
RESPONSE_FLAG_ACKNOWLEDGED = RESPONSE_FLAG_DEFERRED | RESPONSE_FLAG_RESPONDED
RESPONSE_FLAG_DEFERRING_OR_DEFERRED = RESPONSE_FLAG_DEFERRING | RESPONSE_FLAG_DEFERRED
RESPONSE_FLAG_RESPONDING_OR_RESPONDED = RESPONSE_FLAG_RESPONDING | RESPONSE_FLAG_RESPONDED
RESPONSE_FLAG_ACKNOWLEDGING_OR_ACKNOWLEDGED = RESPONSE_FLAG_ACKNOWLEDGING | RESPONSE_FLAG_ACKNOWLEDGED

INTERACTION_TYPE_APPLICATION_COMMAND = InteractionType.application_command


class ApplicationCommandInteraction(DiscordEntity):
    """
    Represents an ``ApplicationCommand`` invoked by a user.
    
    Attributes
    ----------
    id : `int`
        The represented application command's identifier number.
    name : `str`
        The name of the command. It's length can be in range [1:32].
    options : `None`, `tuple` of ``ApplicationCommandInteractionOption``
        The parameters and values from the user if any. Defaults to `None` if non is received.
    resolved_attachments : `None`, `dict` of (`int`, `Attachment``) items
        Resolved received attachments stored by their identifier as keys if any.
    resolved_channels : `None`, `dict` of (`int`, ``ChannelBase``) items
        Resolved received channels stored by their identifier as keys if any.
    resolved_roles : `None`, `dict` of (`int`, ``Role``) items
        Resolved received roles stored by their identifier as keys if any.
    resolved_messages : `None`, `dict` of (`int`, ``Message``) items
        Resolved received messages stored by their identifier as keys if any.
    resolved_users : `None`, `dict` of (`int`, ``ClientUserBase``) items
        Resolved received users stored by their identifier as keys if any.
    target_id : `int`
        The interaction's target's identifier.
    """
    __slots__ = (
        'name', 'options', 'resolved_attachments', 'resolved_channels', 'resolved_roles', 'resolved_messages',
        'resolved_users', 'target_id'
    )
    
    def __new__(cls, data, interaction_event):
        """
        Creates a new ``ApplicationCommandInteraction`` from the data received from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            The received application command interaction data.
        interaction_event : ``InteractionEvent``
            The parent interaction event.
        """
        # id
        id_ = int(data['id'])
        
        # name
        name = data['name']
        
        # options
        option_datas = data.get('options', None)
        if (option_datas is None) or (not option_datas):
            options = None
        else:
            options = tuple(ApplicationCommandInteractionOption(option_data) for option_data in option_datas)
        
        # resolved_channels & resolved_roles & resolved_messages & resolved_users
        try:
            resolved_data = data['resolved']
        except KeyError:
            resolved_attachments = None
            resolved_channels = None
            resolved_roles = None
            resolved_messages = None
            resolved_users = None
        else:
            # resolved_attachments
            try:
                resolved_attachment_datas = resolved_data['attachments']
            except KeyError:
                resolved_attachments = None
            else:
                if resolved_attachment_datas:
                    resolved_attachments = {}
                    
                    for attachment_data in resolved_attachment_datas.values():
                        attachment = Attachment(attachment_data)
                        if (attachment is not None):
                            resolved_attachments[attachment.id] = attachment
                    
                    if not resolved_attachments:
                        resolved_attachments = None
                else:
                    resolved_attachments = None
            
            # resolved_channels
            try:
                resolved_channel_datas = resolved_data['channels']
            except KeyError:
                resolved_channels = None
            else:
                if resolved_channel_datas:
                    resolved_channels = {}
                    
                    for channel_data in resolved_channel_datas.values():
                        channel = create_partial_channel_from_data(channel_data, interaction_event.id)
                        if (channel is not None):
                            resolved_channels[channel.id] = channel
                    
                    if not resolved_channels:
                        resolved_channels = None
                else:
                    resolved_channels = None
            
            # resolved_roles
            try:
                resolved_role_datas = resolved_data['roles']
            except KeyError:
                resolved_roles = None
            else:
                if resolved_role_datas:
                    resolved_roles = {}
                    for role_data in resolved_role_datas.values():
                        role = Role(role_data, interaction_event.guild)
                        resolved_roles[role.id] = role
                else:
                    resolved_roles = None
            
            # resolved_messages
            try:
                resolved_message_datas = resolved_data['messages']
            except KeyError:
                resolved_messages = None
            else:
                if resolved_message_datas:
                    resolved_messages = {}
                    
                    for message_data in resolved_message_datas.values():
                        message = Message(message_data)
                        resolved_messages[message.id] = message
                else:
                    resolved_messages = None
            
            # resolved_users
            try:
                resolved_user_datas = resolved_data['users']
            except KeyError:
                resolved_users = None
            else:
                if resolved_user_datas:
                    try:
                        resolved_guild_profile_datas = resolved_data['members']
                    except KeyError:
                        resolved_guild_profile_datas = None
                    
                    resolved_users = {}
                    
                    for user_id, user_data in resolved_user_datas.items():
                        if resolved_guild_profile_datas is None:
                            guild_profile_data = None
                        else:
                            guild_profile_data = resolved_guild_profile_datas.get(user_id, None)
                        
                        if (guild_profile_data is not None):
                            user_data['member'] = guild_profile_data
                        
                        user = User(user_data, interaction_event.guild)
                        resolved_users[user.id] = user
                        
                        if (guild_profile_data is not None):
                            interaction_event._add_cached_user(user)
                
                else:
                    resolved_users = None
        
        # target_id
        target_id = data.get('target_id', None)
        if target_id is None:
            target_id = 0
        else:
            target_id = int(target_id)
        
        self = object.__new__(cls)
        self.id = id_
        self.name = name
        self.options = options
        self.resolved_attachments = resolved_attachments
        self.resolved_users = resolved_users
        self.resolved_channels = resolved_channels
        self.resolved_roles = resolved_roles
        self.resolved_messages = resolved_messages
        self.target_id = target_id
        
        interaction_event._add_response_waiter()
        
        return self
    
    
    def __repr__(self):
        """Returns the application command interaction's representation."""
        repr_parts = [
            '<', self.__class__.__name__,
            ' id=', repr(self.id),
            ', name=', repr(self.name),
        ]
        
        options = self.options
        if (options is not None):
            repr_parts.append(', options=[')
            
            index = 0
            limit = len(options)
            
            while True:
                option = options[index]
                index += 1
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        
        target = self.target
        if (target is not None):
            repr_parts.append(', target=')
            repr_parts.append(repr(target))
        
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two application command interactions are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # id
        if self.id != other.id:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # options
        if self.options != other.options:
            return False
        
        # resolved_attachments
        if self.resolved_attachments != other.resolved_attachments:
            return False
        
        # resolved_channels
        if self.resolved_channels != other.resolved_channels:
            return False
        
        # resolved_roles
        if self.resolved_roles != other.resolved_roles:
            return False
        
        # resolved_messages
        if self.resolved_messages != other.resolved_messages:
            return False
        
        # resolved_users
        if self.resolved_users != other.resolved_users:
            return False
        
        # target_id
        if self.target_id != other.target_id:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the hash value of the application command interaction."""
        hash_value = 0
        
        # id
        hash_value ^= self.id
        
        # name
        hash_value ^= hash(self.name)
        
        # options
        options = self.options
        if (options is not None):
            hash_value ^= len(options)
            
            for option in options:
                hash_value ^= hash(option)
        
        # resolved_attachments
        resolved_attachments = self.resolved_attachments
        if (resolved_attachments is not None):
            hash_value ^= (len(resolved_attachments) << 24)
            
            for attachment_id in resolved_attachments.keys():
                hash_value ^= attachment_id
        
        # resolved_channels
        resolved_channels = self.resolved_channels
        if (resolved_channels is not None):
            hash_value ^= (len(resolved_channels) << 8)
            
            for channel_id in resolved_channels.keys():
                hash_value ^= channel_id
        
        # resolved_roles
        resolved_roles = self.resolved_roles
        if (resolved_roles is not None):
            hash_value ^= (len(resolved_roles) << 12)
            
            for role_id in resolved_roles.keys():
                hash_value ^= role_id
        
        # resolved_messages
        resolved_messages = self.resolved_messages
        if (resolved_messages is not None):
            hash_value ^= (len(resolved_messages) << 16)
            
            for message_id in resolved_messages.keys():
                hash_value ^= message_id
        
        # resolved_users
        resolved_users = self.resolved_users
        if (resolved_users is not None):
            hash_value ^= (len(resolved_users) << 20)
            
            for user_id in resolved_users.keys():
                hash_value ^= user_id
        
        # target_id
        hash_value ^= self.target_id
        
        return hash_value
    
    
    def resolve_entity(self, entity_id):
        """
        Tries to resolve the entity by the given identifier.
        
        Parameters
        ----------
        entity_id : ``int``
            The entity's identifier.
        
        Returns
        -------
        resolved : `None`, ``DiscordEntity``
            The resolved discord entity if found.
        """
        # Is used at `InteractionEvent.target`, which wanna access user and message first, so we check that two first.
        resolved_messages = self.resolved_messages
        if (resolved_messages is not None):
            try:
                entity = resolved_messages[entity_id]
            except KeyError:
                pass
            else:
                return entity
        
        
        resolved_users = self.resolved_users
        if (resolved_users is not None):
            try:
                entity = resolved_users[entity_id]
            except KeyError:
                pass
            else:
                return entity
        
        
        resolved_roles = self.resolved_roles
        if (resolved_roles is not None):
            try:
                entity = resolved_roles[entity_id]
            except KeyError:
                pass
            else:
                return entity
        
        
        resolved_channels = self.resolved_channels
        if (resolved_channels is not None):
            try:
                entity = resolved_channels[entity_id]
            except KeyError:
                pass
            else:
                return entity
        
        
        resolved_attachments = self.resolved_attachments
        if (resolved_attachments is not None):
            try:
                entity = resolved_attachments[entity_id]
            except KeyError:
                pass
            else:
                return entity
        
        
        return None
    
    @property
    def target(self):
        """
        Returns the interaction event's target.
        
        Only applicable for context application commands.
        
        Returns
        -------
        target : ``ClientUserBase``, ``Message``
        """
        target_id = self.target_id
        if target_id:
            return self.resolve_entity(target_id)


class ApplicationCommandInteractionOption:
    """
    Represents an option of a ``ApplicationCommandInteraction``.
    
    Attributes
    ----------
    name : `str`
        The option's name.
    
    options : `None`, `tuple` of ``ApplicationCommandInteractionOption``
        The parameters and values from the user. Present if a sub-command was used. Defaults to `None` if non is
        received.
        
        Mutually exclusive with the `value` field.
    
    type : ``ApplicationCommandOptionType``
        The option's type.
    
    value : `None`, `str`, `bool`, `float`, `int`
        The given value by the user.
        
        Mutually exclusive with the `options` field,
    
    """
    __slots__ = ('name', 'options', 'type', 'value')
    
    def __new__(cls, data):
        """
        Creates a new ``ApplicationCommandInteractionOption`` from the data received from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            The received application command interaction option data.
        """
        # name
        name = data['name']
        
        # options
        option_datas = data.get('options', None)
        if (option_datas is None) or (not option_datas):
            options = None
        else:
            options = tuple(ApplicationCommandInteractionOption(option_data) for option_data in option_datas)
        
        # type
        type_ = ApplicationCommandOptionType.get(data.get('type', 0))
        
        # value
        value = data.get('value', None)
        
        self = object.__new__(cls)
        
        self.name = name
        self.options = options
        self.type = type_
        self.value = value
        
        return self
    
    
    def __repr__(self):
        """Returns the application command interaction option's representation."""
        repr_parts = [
            '<', self.__class__.__name__,
            ', name=', repr(self.name),
        ]
        
        type_ = self.type
        if type_ is not ApplicationCommandOptionType.none:
            repr_parts.append(', type=')
            repr_parts.append(type_.name)
            repr_parts.append(' (')
            repr_parts.append(repr(type_.value))
            repr_parts.append(')')
        
        options = self.options
        if (options is not None):
            repr_parts.append(', options=[')
            
            index = 0
            limit = len(options)
            
            while True:
                option = options[index]
                index += 1
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        value = self.value
        if (value is not None):
            repr_parts.append(', value=')
            repr_parts.append(repr(value))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two application command interaction options are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # name
        if self.name != other.name:
            return False
        
        # options
        if self.options != other.options:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        # value
        if self.value != other.value:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the application command interaction option's hash value."""
        hash_value = 0
        
        # name
        hash_value ^= hash(self.name)
        
        # options
        options = self.options
        if (options is not None):
            hash_value ^= len(options)
            
            for option in options:
                hash_value ^= hash(option)
        
        # type
        hash_value ^= (self.type.value << 8)
        
        # value
        value = self.value
        if (value is not None):
            hash_value ^= hash(value)
        
        return hash_value


class ComponentInteraction:
    """
    A component interaction of an ``InteractionEvent``.
    
    Attributes
    ----------
    custom_id : `None`, `str`
        The component's custom identifier.
    options : `None`, `tuple` of `str`
        Option values selected of the respective interaction.
    type : ``ComponentType``
        The component's type.
    """
    __slots__ = ('type', 'custom_id', 'components', 'options')
    
    def __new__(cls, data, interaction_event):
        """
        Creates a new component interaction with the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            The received application command interaction data.
        interaction_event : ``InteractionEvent``
            The parent interaction event.
        """
        # custom_id
        custom_id = data.get('custom_id', None)
        
        # type
        type_ = ComponentType.get(data['component_type'])
        
        # options
        option_datas = data.get('values', None)
        if (option_datas is None) or (not option_datas):
            options = None
        else:
            options = tuple(option_datas)
        
        self = object.__new__(cls)
        self.custom_id = custom_id
        self.options = options
        self.type = type_
        return self
    
    
    def __repr__(self):
        """Returns the component interaction's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # type
        type_ = self.type
        repr_parts.append(' type=')
        repr_parts.append(type_.name)
        repr_parts.append(' (')
        repr_parts.append(repr(type_.value))
        repr_parts.append(')')
        
        custom_id = self.custom_id
        if (custom_id is not None):
            repr_parts.append(', custom_id=')
            repr_parts.append(reprlib.repr(custom_id))
        
        options = self.options
        if (options is not None):
            repr_parts.append(', options=[')
            index = 0
            limit = len(options)
            while True:
                option = options[index]
                repr_parts.append(repr(option))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Compares the two component or component interaction."""
        other_type = type(other)
        # Compare with self type.
        if other_type is type(self):
            # type
            if self.type is not other.type:
                return False
            #custom_id
            if self.custom_id != other.custom_id:
                return False
            
            return True
        
        # Compare with components.
        if issubclass(other_type, ComponentBase):
            # Check `type` before `custom_id`
            
            # type
            if self.type is not other.type:
                return False
            
            # custom_id
            if self.custom_id != other.custom_id:
                return False
            
            return True
        
        
        return NotImplemented
    
    
    def __hash__(self):
        """Returns the component interaction's hash value."""
        hash_value = 0
        
        # type
        hash_value ^= self.type.value
        
        # custom_id
        hash_value ^= hash(self.custom_id)
        
        # options
        options = self.options
        if (options is not None):
            hash_value ^ len(options) << 8
            for option in options:
                hash_value ^ hash(option)
        
        return hash_value
    
    
    def component_type(cls):
        """
        ``.component_type`` is deprecated, please use ``.type`` instead. Will be removed in 2022
        February.
        """
        warnings.warn(
            f'`{cls.__name__}.component_type` is deprecated, and will be removed in 2022 February. '
            f'Please use `{cls.__name__}.type` instead.',
            FutureWarning)
        
        return cls.type


class ApplicationCommandAutocompleteInteractionOption:
    """
    Application auto complete option representing an auto completable parameters.
    
    Attributes
    ----------
    focused : `bool`
        Whether this field is focused by the user.
    name : `str`
        The name of the parameter.
    options : `None`, `tuple` of ``ApplicationCommandAutocompleteInteractionOption``
        Nested functions.
    type : ``ApplicationCommandOptionType``
        The represented option's type.
    value : `None`, `str`
        The value, the user has been typed.
    """
    __slots__ = ('focused', 'name', 'options', 'type', 'value')
    
    def __new__(cls, data):
        """
        Creates a new ``ApplicationCommandAutocompleteOption`` from the data received from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`)
            Application command autocomplete option data.
        """
        # focused
        focused = data.get('focused', False)
        
        # name
        name = data['name']
        
        # options
        option_datas = data.get('options', None)
        if (option_datas is None) or (not option_datas):
            options = None
        else:
            options = tuple(
                ApplicationCommandAutocompleteInteractionOption(option_data) for option_data in option_datas
            )
        
        # type
        type_ = ApplicationCommandOptionType.get(data['type'])
        
        # value
        value = data.get('value', None)
        if (value is not None) and (not value):
            value = None
        
        self = object.__new__(cls)
        self.focused = focused
        self.name = name
        self.options = options
        self.type = type_
        self.value = value
        return self
    
    
    def __repr__(self):
        """Returns the application command autocomplete option's representation."""
        repr_parts = [
            '<', self.__class__.__name__,
            ' name=', repr(self.name),
        ]
        
        if self.focused:
            repr_parts.append(' (focused)')
        
        value = self.value
        if (value is not None):
            repr_parts.append(', value=')
            repr_parts.append(reprlib.repr(value))
        
        type_ = self.type
        repr_parts.append(', type=')
        repr_parts.append(type_.name)
        repr_parts.append(' (')
        repr_parts.append(repr(type_.value))
        repr_parts.append(')')
        
        options = self.options
        if (options is not None):
            repr_parts.append(', options=[')
            
            index = 0
            limit = len(options)
            
            while True:
                option = options[index]
                index += 1
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the application command autocomplete option's representation."""
        hash_value = 0
        
        # focused
        hash_value ^= (self.focused << 16)
        
        # name
        hash_value ^= hash(self.name)
        
        # options
        options = self.options
        if (options is not None):
            hash_value ^= (len(options) << 8)
            
            for option in options:
                hash_value ^= hash(option)
        
        # type
        hash_value ^= self.type.value
        
        # value
        value = self.value
        if (value is not None):
            hash_value ^= hash(value)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two application command autocomplete option's are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # focused
        if self.focused != other.focused:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # options
        if self.options != other.options:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        # value
        if self.value != other.value:
            return False
        
        return True
    
    
    @property
    def focused_option(self):
        """
        Returns the focused option of the application command autocomplete interaction option.
        
        Returns
        -------
        option : `None`, ``ApplicationCommandAutocompleteInteractionOption``
        """
        if self.focused:
            return self
        
        options = self.options
        if (options is not None):
            for option in options:
                focused_option =  option.focused_option
                if (focused_option is not None):
                    return focused_option
    
    
    def get_value_of(self, *option_names):
        """
        Gets the value for the option by the given name.
        
        Parameters
        ----------
        *option_names : `str`
            The option(s)'s name.
        
        Returns
        -------
        value : `None`, `str`
            The value, the user has been typed.
        """
        if option_names:
            option_name, *option_names = option_names
            
            options = self.options
            if options is None:
                value = None
            else:
                for option in options:
                    if option.name == option_name:
                        value = option.get_value_of(*option_names)
                        break
                else:
                    value = None
        else:
            value = self.value
        
        return value


class ApplicationCommandAutocompleteInteraction:
    """
    Represents an ``ApplicationCommand``'s auto completion interaction.
    
    Attributes
    ----------
    id : `int`
        The represented application command's identifier number.
    name : `str`
        The name of the command. It's length can be in range [1:32].
    options : `None`, `tuple` of ``ApplicationCommandAutocompleteOption``
        Parameter auto completion options.
    """
    __slots__ = ('id', 'name', 'options',)
    
    def __new__(cls, data, interaction_event):
        """
        Creates a new ``ApplicationCommandAutocompleteInteraction`` from the data received from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            The received application command interaction data.
        interaction_event : ``InteractionEvent``
            The parent interaction event.
        """
        # id
        id_ = int(data['id'])
        
        # name
        name = data['name']
        
        # options
        option_datas = data.get('options', None)
        if (option_datas is None) or (not option_datas):
            options = None
        else:
            options = tuple(
                ApplicationCommandAutocompleteInteractionOption(option_data) for option_data in option_datas
            )
        
        self = object.__new__(cls)
        
        self.id = id_
        self.name = name
        self.options = options
        
        return self
    
    
    def __repr__(self):
        """Returns the application command interaction auto completion's representation."""
        repr_parts = [
            '<', self.__class__.__name__,
            ' id=', repr(self.id),
            ', name=', repr(self.name),
        ]
        
        options = self.options
        if (options is not None):
            repr_parts.append(', options=[')
            
            index = 0
            limit = len(options)
            
            while True:
                option = options[index]
                index += 1
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the application command autocomplete interaction hash value."""
        hash_value = 0
        
        # id
        hash_value ^= self.id
        
        # name
        hash_value ^= hash(self.name)
        
        # options
        options = self.options
        if (options is not None):
            hash_value ^= (len(options) << 8)
            
            for option in options:
                hash_value ^= hash(option)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two application command autocomplete interaction are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # id
        if self.id != other.id:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # options
        if self.options != other.options:
            return False
        
        return True
        
    
    @property
    def focused_option(self):
        """
        Returns the focused option of the application command autocomplete interaction.
        
        Returns
        -------
        option : `None`, ``ApplicationCommandAutocompleteInteractionOption``
        """
        options = self.options
        if options is None:
            focused_option = None
        else:
            for option in options:
                focused_option = option.focused_option
                if (focused_option is not None):
                    break
            else:
                focused_option = None
        
        return focused_option
    
    
    def get_value_of(self, *option_names):
        """
        Gets the value for the option by the given name.
        
        Parameters
        ----------
        *option_names : `str`
            The option(s)'s name.
        
        Returns
        -------
        value : `None`, `str`
            The value, the user has been typed.
        """
        if option_names:
            option_name, *option_names = option_names
            
            options = self.options
            if options is None:
                value = None
            else:
                for option in options:
                    if option.name == option_name:
                        value = option.get_value_of(*option_names)
                        break
                else:
                    value = None
        else:
            value = None
        
        return value
    
    
    @property
    def value(self):
        """
        Returns the focused option's value of the application command autocomplete interaction.
        
        Returns
        -------
        value : `None`, `str`
        """
        focused_option = self.focused_option
        if (focused_option is None):
            value = None
        else:
            value = focused_option.value
        
        return value


class FormSubmitInteraction:
    """
    Represents a response to a ``InteractionForm``.
    
    Attributes
    ----------
    custom_id : `None`, `str`
        The forms's custom identifier.
    options : `None`, `tuple` of ``FormSubmitInteractionOption``
        Submitted component values.
    """
    __slots__ = ('custom_id', 'options', )
    
    def __new__(cls, data, interaction_event):
        """
        Creates a new component interaction with the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            The received form submit interaction data.
        interaction_event : ``InteractionEvent``
            The parent interaction event.
        """
        # custom_id
        custom_id = data.get('custom_id', None)
        if (custom_id is not None) and (not custom_id):
            custom_id = None
        
        # options
        option_datas = data.get('components', None)
        if (option_datas is None) or (not option_datas):
            options = None
        else:
            options = tuple(FormSubmitInteractionOption(option_data) for option_data in option_datas)
        
        
        self = object.__new__(cls)
        
        self.custom_id = custom_id
        self.options = options
        
        return self


    def __repr__(self):
        """Returns the form submit interaction's representation."""
        repr_parts = ['<', self.__class__.__name__,]
        
        repr_parts.append('custom_id=')
        repr_parts.append(self.custom_id)
        
        options = self.options
        if (options is not None):
            repr_parts.append(', options=[')
            
            index = 0
            limit = len(options)
            
            while True:
                option = options[index]
                index += 1
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the form submit interaction's hash value."""
        hash_value = 0
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            hash_value ^= hash(custom_id)
        
        # options
        options = self.options
        if (options is not None):
            hash_value ^= (len(options) << 8)
            
            for option in options:
                hash_value ^= hash(option)
            
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two submit interactions are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # custom_id
        if self.custom_id != other.custom_id:
            return False
        
        # options
        if self.options != other.options:
            return False
        
        return True
    
    
    def iter_custom_ids_and_values(self):
        """
        Iterates over all the `custom_id`-s and values of the form submit interaction.
        
        This method is an iterable generator.
        
        Yields
        ------
        custom_id : `str`
            The `custom_id` of a represented component.
        value : `str`
            The `value` passed by the user.
        """
        options = self.options
        if (options is not None):
            for option in options:
                yield from option.iter_custom_ids_and_values()
    
    
    def get_custom_id_value_relation(self):
        """
        Returns a dictionary with `custom_id` to `value` relation.
        
        Returns
        -------
        custom_id_value_relation : `dict` of (`str`, `str`) items
        """
        custom_id_value_relation = {}
        
        for custom_id, value in self.iter_custom_ids_and_values():
            if (value is not None):
                custom_id_value_relation[custom_id] = value
        
        return custom_id_value_relation
    
    
    def get_value_for(self, custom_id_to_match):
        """
        Returns the value for the given `custom_id`.
        
        Parameters
        ----------
        custom_id_to_match : `str`
            A respective components `custom_id` to match.
        
        Returns
        -------
        value : `None`, `str`
            The value if any.
        """
        for custom_id, value in self.iter_custom_ids_and_values():
            if (custom_id == custom_id):
                return value
    
    
    def get_match_and_value(self, matcher):
        """
        Gets a `custom_id`'s value matching the given `matcher`.
        
        Parameters
        ----------
        matcher : `callable`
            Matcher to call on a `custom_id`
            
            Should accept the following parameters:
            
            +-----------+-----------+
            | Name      | Type      |
            +===========+===========+
            | custom_id | `str`     |
            +-----------+-----------+
            
            Should return non-`None` on success.
        
        Returns
        -------
        match : `None`, `Any`
            The returned value by the ``matcher``
        value : `None`, `str`
            The matched `custom_id`'s value.
        """
        for custom_id, value in self.iter_custom_ids_and_values():
            match = matcher(custom_id)
            if (match is not None):
                return match, value
        
        return None, None
    
    
    def iter_matches_and_values(self, matcher):
        """
        Gets a `custom_id`'s value matching the given `matcher`.
        
        This method is an iterable generator.
        
        Parameters
        ----------
        matcher : `callable`
            Matcher to call on a `custom_id`
            
            Should accept the following parameters:
            
            +-----------+-----------+
            | Name      | Type      |
            +===========+===========+
            | custom_id | `str`     |
            +-----------+-----------+
            
            Should return non-`None` on success.
        
        Yields
        -------
        match : `None`, `Any`
            The returned value by the ``matcher``
        value : `None`, `str`
            The matched `custom_id`'s value.
        """
        for custom_id, value in self.iter_custom_ids_and_values():
            match = matcher(custom_id)
            if (match is not None):
                yield match, value


class FormSubmitInteractionOption:
    """
    Attributes
    ----------
    custom_id : `None`, `str`
        The option's respective component's type.
        
    options : `None`, `tuple` of ``FormSubmitInteractionOption``
        Mutually exclusive with the `value` field.
    
    type : ``ComponentType``
        The option respective component's type.
        
    value : `None`, `str`
        Mutually exclusive with the `options` field.
    """
    __slots__ = ('custom_id', 'options', 'type', 'value')
    
    def __new__(cls, data):
        """
        Creates a new ``FormSubmitInteractionOption`` from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            The received form submit interaction option data.
        """
        # custom_id
        custom_id = data.get('custom_id', None)
        if (custom_id is not None) and (not custom_id):
            custom_id = None
        
        # options
        option_datas = data.get('components', None)
        if (option_datas is None) or (not option_datas):
            options = None
        else:
            options = tuple(FormSubmitInteractionOption(option_data) for option_data in option_datas)
        
        # type
        type_ = ComponentType.get(data.get('type', 0))
        
        # value
        value = data.get('value', None)
        if (value is not None) and (not value):
            value = None
        
        self = object.__new__(cls)
        
        self.custom_id = custom_id
        self.options = options
        self.type = type_
        self.value = value
        
        return self
        
    def __repr__(self):
        """Returns the application command interaction option's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # Descriptive fields : type
        
        # type
        type_ = self.type
        if type_ is not ComponentType.none:
            repr_parts.append(', type=')
            repr_parts.append(type_.name)
            repr_parts.append(' (')
            repr_parts.append(repr(type_.value))
            repr_parts.append(')')
        
        # System fields : custom_id
        
        # custom_id
        repr_parts.append(', custom_id=')
        repr_parts.append(reprlib.repr(self.custom_id))
        
        # Extra descriptive fields : options | value
        # options
        options = self.options
        if (options is not None):
            repr_parts.append(', options=[')
            
            index = 0
            limit = len(options)
            
            while True:
                option = options[index]
                index += 1
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        # value
        value = self.value
        if (value is not None):
            repr_parts.append(', value=')
            repr_parts.append(repr(value))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)


    def __hash__(self):
        """Returns the form submit interaction's option hash value."""
        hash_value = 0
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            hash_value ^= hash(custom_id)
        
        # options
        options = self.options
        if (options is not None):
            hash_value ^= (len(options) << 8)
            
            for option in options:
                hash_value ^= hash(option)
        
        # type
        hash_value ^= self.type.value
        
        # value
        value = self.value
        if (value is not None):
            hash_value ^= hash(value)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two submit interaction options are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # custom_id
        if self.custom_id != other.custom_id:
            return False
        
        # options
        if self.options != other.options:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        # value
        if self.value != other.value:
            return False
        
        return True
    
    
    def iter_custom_ids_and_values(self):
        """
        Iterates over all the `custom_id`-s and values of the form submit interaction option.
        
        This method is an iterable generator.
        
        Yields
        ------
        custom_id : `str`
            The `custom_id` of a represented component.
        value : `str`
            The `value` passed by the user.
        """
        custom_id = self.custom_id
        if (custom_id is not None):
            yield custom_id, self.value
        
        options = self.options
        if (options is not None):
            for option in options:
                yield from option.iter_custom_ids_and_values()


INTERACTION_TYPE_TABLE = {
    InteractionType.ping.value: None,
    InteractionType.application_command.value: ApplicationCommandInteraction,
    InteractionType.message_component.value: ComponentInteraction,
    InteractionType.application_command_autocomplete.value: ApplicationCommandAutocompleteInteraction,
    InteractionType.form_submit: FormSubmitInteraction,
}


@export
class InteractionEvent(DiscordEntity, EventBase, immortal=True):
    """
    Represents a processed `INTERACTION_CREATE` dispatch event.
    
    Attributes
    ----------
    id : `int`
        The interaction's id.
    _cached_users : `None`, `list` of ``ClientUserBase``
        A list of users, which are temporary cached.
    _response_flag : `bool`
        The response order state of ``InteractionEvent``
        
        +-------------------------------+-------+---------------------------------------------------+
        | Respective name               | Shift | Description                                       |
        +===============================+=======+===================================================+
        | RESPONSE_FLAG_DEFERRING       | 0     | The vent is being acknowledged.                   |
        +-------------------------------+-------+---------------------------------------------------+
        | RESPONSE_FLAG_DEFERRED        | 1     | The event was acknowledged and response will be   |
        |                               |       | sent later. Shows loading screen for the user.    |
        +-------------------------------+-------+---------------------------------------------------+
        | RESPONSE_FLAG_RESPONDING      | 2     | Responding to the interaction.                    |
        +-------------------------------+-------+---------------------------------------------------+
        | RESPONSE_FLAG_RESPONDED       | 3     | Response was sent on the interaction.             |
        +-------------------------------+-------+---------------------------------------------------+
        | RESPONSE_FLAG_EPHEMERAL       | 4     | Whether the main response is an ephemeral,        |
        |                               |       | showing for the invoking user only.               |
        +-------------------------------+-------+---------------------------------------------------+
        
        Can be used by extensions and is used by the the ``Client``-s to ensure correct flow order.
    application_id : `int`
        The interaction's application's identifier.
    channel_id : `int`
        The channel's identifier from where the interaction was called.
    guild_id : `int`
        The guild's identifier from where the interaction was called from. Might be `0` if the interaction was called
        from a private channel.
    guild_locale : `str`
        The guild's preferred locale if invoked from guild.
    interaction : `None`, ``ApplicationCommandInteraction``, ``ComponentInteraction`` or \
            ``ApplicationCommandAutocompleteInteraction``
        
        The called interaction by it's route by the user.
    locale : `str`
        The selected language of the invoking user.
    message : `None`, ``Message``
        The message from where the interaction was received. Applicable for message components.
    token : `str`
        Interaction's token used when responding on it.
    type : ``InteractionType``
        The interaction's type.
    user : ``ClientUserBase``
        The user who called the interaction.
    user_permissions : ``Permission``
        The user's permissions in the respective channel.
    
    Class Attributes
    ----------------
    _USER_GUILD_CACHE : `dict` of (`tuple` (``ClientUserBase``, ``Guild``), `int`)
        A cache which stores `user-guild` pairs as keys and their reference count as values to remember
        ``InteractionEvent``'s ``.user``-s' guild profiles of the respective ``.guild`` even if the ``Guild`` is
        uncached.
    
        Note, that private channel interaction, neither interactions of cached guilds are not added, what means if
        all the clients are kicked from a guild the guild profile can be lost in unexpected time.
    
    Notes
    -----
    The interaction token can be used for 15 minutes, tho if it is not used within the first 3 seconds, it is
    invalidated immediately.
    
    ˙˙InteractionEvent˙˙ instances are weakreferable.
    """
    __slots__ = (
        '_cached_users', '_response_flag', 'application_id', 'channel_id', 'guild_id', 'guild_locale', 'interaction',
        'locale', 'message', 'token', 'type', 'user', 'user_permissions'
    )
    
    _USER_GUILD_CACHE = {}
    
    def __new__(cls, data):
        """
        Creates a new ``InteractionEvent`` with the given parameters.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            `INTERACTION_CREATE` dispatch event data.
        """
        # id_
        id_ = int(data['id'])
        
        # application_id
        application_id = int(data['application_id'])
        
        # channel_id
        channel_id = int(data['channel_id'])
        
        # guild_id
        guild_id = data.get('guild_id', None)
        if guild_id is None:
            guild_id = 0
        else:
            guild_id = int(guild_id)
        
        if guild_id:
            guild = create_partial_guild_from_id(guild_id)
        else:
            guild = None
        
        # guild_locale
        guild_locale = parse_guild_locale(data)
        
        # interaction
        # We set interaction at the end when the object is fully initialized
        
        # locale
        locale = parse_locale(data)
        
        # message
        try:
            message_data = data['message']
        except KeyError:
            message = None
        else:
            message = Message(message_data)
        
        # token
        token = data['token']
        
        # type
        type_value = data['type']
        type_ = InteractionType.get(type_value)
        
        # user
        try:
            user_data = data['member']
        except KeyError:
            user_data = data['user']
        
        user = User(user_data, guild)
        
        # user_permissions
        try:
            user_permissions = user_data['permissions']
        except KeyError:
            user_permissions = PERMISSION_PRIVATE
        else:
            user_permissions = Permission(user_permissions)
        
        
        self = object.__new__(cls)
        self.id = id_
        self.application_id = application_id
        self.type = type_
        self.channel_id = channel_id
        self.guild_id = guild_id
        self.guild_locale = guild_locale
        self.interaction = None
        self.locale = locale
        self.token = token
        self.user = user
        self.user_permissions = user_permissions
        self._response_flag = RESPONSE_FLAG_NONE
        self._cached_users = None
        self.message = message
        
        # Cache our user if called from guild. This is required to kill references to the user in the guild.
        if (guild is not None):
            self._add_cached_user(user)
        
        # All field is set -> we can now create our own child
        interaction_type = INTERACTION_TYPE_TABLE.get(type_value, None)
        if (interaction_type is not None):
            self.interaction = interaction_type(data['data'], self)
        
        # Bind cached users to the guild for un-caching on object unallocation.
        cached_users = self._cached_users
        if (cached_users is not None):
            for user in cached_users:
                key = (user, guild)
                USER_GUILD_CACHE = cls._USER_GUILD_CACHE
                try:
                    reference_count = USER_GUILD_CACHE[key]
                except KeyError:
                    reference_count = 1
                else:
                    reference_count += 1
                
                USER_GUILD_CACHE[key] = reference_count
        
        return self
    
    
    async def wait_for_response_message(self, *, timeout=None):
        """
        Waits for response message. Applicable for application command interactions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        timeout : `None`, `float`, Optional (Keyword only)
            The maximal time to wait for message before `TimeoutError` is raised.
        
        Returns
        -------
        message : ``Message``
            The received message.
        
        Raises
        ------
        RuntimeError
            The interaction was acknowledged with `show_for_invoking_user_only=True` (as ephemeral). Response
            `message` cannot be detected.
        TimeoutError
            Message was not received before timeout.
        """
        message = self.message
        if (message is not None):
            return message
        
        if self._response_flag & RESPONSE_FLAG_EPHEMERAL:
            raise RuntimeError(
                f'The interaction was acknowledged with `show_for_invoking_user_only=True` '
                f'(as ephemeral). Response `message` cannot be detected.'
            )
        
        try:
            waiter = INTERACTION_EVENT_MESSAGE_WAITERS[self]
        except KeyError:
            waiter = Future(KOKORO)
            INTERACTION_EVENT_MESSAGE_WAITERS[self] = waiter
        
        waiter = shield(waiter, KOKORO)
        
        if (timeout is not None):
            future_or_timeout(waiter, timeout)
        
        await waiter
        return self.message
    
    
    def __del__(self):
        """
        Unregisters the user-guild pair from the interaction cache.
        """
        cached_users = self._cached_users
        if cached_users is None:
            return
        
        guild = self.guild
        if (guild is None):
            return
        
        for user in cached_users:
            key = (user, guild)
            USER_GUILD_CACHE = self._USER_GUILD_CACHE
            
            # A client meanwhile joined the guild?
            if not guild.partial:
                try:
                    del USER_GUILD_CACHE[key]
                except KeyError:
                    pass
                return
            
            try:
                reference_count = USER_GUILD_CACHE[key]
            except KeyError:
                reference_count = 0
            else:
                if reference_count == 1:
                    del USER_GUILD_CACHE[key]
                    reference_count = 0
                else:
                    reference_count -= 1
            
            if reference_count == 0:
                try:
                    del user.guild_profiles[guild.id]
                except KeyError:
                    pass
    
    
    def __repr__(self):
        """Returns the representation of the event."""
        repr_parts = ['<', self.__class__.__name__]
        
        response_state_names = None
        response_state = self._response_flag
        
        if response_state == RESPONSE_FLAG_NONE:
            pass
        
        if response_state & RESPONSE_FLAG_DEFERRING:
            if response_state_names is None:
                response_state_names = []
            response_state_names.append('deferring')
        
        if response_state & RESPONSE_FLAG_DEFERRED:
            if response_state_names is None:
                response_state_names = []
            response_state_names.append('deferred')
        
        if response_state & RESPONSE_FLAG_RESPONDING:
            if response_state_names is None:
                response_state_names = []
            response_state_names.append('responding')
        
        if response_state & RESPONSE_FLAG_RESPONDED:
            if response_state_names is None:
                response_state_names = []
            response_state_names.append('responded')
        
        if response_state & RESPONSE_FLAG_EPHEMERAL:
            if response_state_names is None:
                response_state_names = []
            response_state_names.append('ephemeral')
        
        if (response_state_names is not None):
            repr_parts.append(' (')
            index = 0
            limit = len(response_state_names)
            while True:
                response_state_name = response_state_names[index]
                repr_parts.append(response_state_name)
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(response_state_name)
            repr_parts.append('),')
        
        repr_parts.append(' type=')
        interaction_type = self.type
        repr_parts.append(interaction_type.name)
        repr_parts.append(' (')
        repr_parts.append(repr(interaction_type.value))
        repr_parts.append(')')
        
        
        guild_id = self.guild_id
        if guild_id:
            repr_parts.append(', guild_id=')
            repr_parts.append(repr(guild_id))
        
        
        repr_parts.append(', channel_id=')
        repr_parts.append(repr(self.channel_id))
        
        
        message = self.message
        if (message is not None):
            repr_parts.append(', message=')
            repr_parts.append(repr(message))
        
        
        repr_parts.append(', user=')
        repr_parts.append(repr(self.user))
        
        repr_parts.append(', guild_locale=')
        repr_parts.append(repr(self.guild_locale))
        
        if guild_id:
            repr_parts.append(', locale=')
            repr_parts.append(repr(self.locale))
        
        repr_parts.append(', interaction=')
        repr_parts.append(repr(self.interaction))
        
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def is_unanswered(self):
        """
        Returns whether the event was not acknowledged and not acknowledging either.
        
        Returns
        -------
        is_unanswered : `bool`
        """
        return True if (self._response_flag == RESPONSE_FLAG_NONE) else False
    
    
    def is_acknowledging(self):
        """
        Returns whether the event is being acknowledged.
        
        Returns
        -------
        is_acknowledging : `bool`
        """
        return True if (self._response_flag & RESPONSE_FLAG_ACKNOWLEDGING) else False
    
    
    def is_acknowledged(self):
        """
        Returns whether the event is acknowledged.
        
        Returns
        -------
        is_acknowledged : `bool`
        """
        return True if (self._response_flag & RESPONSE_FLAG_ACKNOWLEDGED) else False
    
    
    def is_deferred(self):
        """
        Returns whether the event is deferred.
        
        Returns
        -------
        is_deferred : `bool`
        """
        response_state = self._response_flag
        if response_state & RESPONSE_FLAG_RESPONDED:
            return False
        
        if response_state & RESPONSE_FLAG_DEFERRED:
            return True
        
        return False
    
    
    def is_responding(self):
        """
        Returns whether the even it being responded.
        
        Returns
        -------
        is_responding : `bool`
        """
        return True if (self._response_flag & RESPONSE_FLAG_RESPONDING) else False
    
    
    def is_responded(self):
        """
        Returns whether was responded.
        
        Returns
        -------
        is_responded : `bool`
        """
        return True if (self._response_flag & RESPONSE_FLAG_RESPONDED) else False
    
    
    def __len__(self):
        """Helper for unpacking if needed."""
        return 3
    
    
    def __iter__(self):
        """
        Unpacks the event.
        
        This method is a generator.
        """
        yield self.type
        yield self.user
        yield self.interaction
    
    
    @property
    def channel(self):
        """
        Returns the interaction's event.
        
        Returns
        -------
        channel : ``ChannelTextBase``, `None`
        """
        return CHANNELS.get(self.channel_id, None)
    
    
    @property
    def guild(self):
        """
        Returns the interaction's guild.
        
        Returns
        -------
        guild : ``Guild``, `None`
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)
    
    
    @property
    def client(self):
        """
        Returns the interaction's client.
        
        Returns
        -------
        client : `Client`
        
        Raises
        ------
        RuntimeError
            Client could not be identified.
        """
        try:
            return APPLICATION_ID_TO_CLIENT[self.application_id]
        except KeyError:
            raise RuntimeError(
                f'Client of {self!r} could not be identified.'
            ) from None
    
    
    @property
    def voice_client(self):
        """
        Returns the voice client of the interaction's client in it's guild.
        
        Returns
        -------
        voice_client : `None`, ``VoiceClient``
        """
        try:
            client = APPLICATION_ID_TO_CLIENT[self.application_id]
        except KeyError:
            voice_client = None
        else:
            guild_id = self.message.guild_id
            if guild_id:
                voice_client = client.voice_clients.get(guild_id, None)
            else:
                voice_client = None
        
        return voice_client
    
    
    def _add_cached_user(self, user):
        """
        Adds a user to the cached ones by the interaction.
        
        This function might be called inside of ``InteractionEvent.__new__`` when initializing it's interaction.
        
        Parameters
        ----------
        user : ``ClientUserBase``
            The user to add to cache.
        """
        cached_users = self._cached_users
        if cached_users is None:
            self._cached_users = [user]
        else:
            if (user not in cached_users):
                cached_users.append(user)
    
    
    def _add_response_waiter(self):
        """
        Adds the interaction event to response waiters.
        
        Called when the interaction is application command one, to resolve it's ``.message`` attribute when created.
        """
        INTERACTION_EVENT_RESPONSE_WAITERS[self.id] = self
    

class InteractionResponseContext:
    """
    Interaction response context manager for managing the interaction's response flag.
    
    Attributes
    ----------
    interaction : ``InteractionEvent``
        The respective interaction event.
    is_deferring : `bool`
        Whether the request just deferring the interaction.
    is_ephemeral : `bool`
        Whether the request is ephemeral.
    """
    __slots__ = ('interaction', 'is_deferring', 'is_ephemeral',)
    
    def __new__(cls, interaction, is_deferring, is_ephemeral):
        """
        Creates a new ``InteractionResponseContext`` with the given parameters.
        
        Parameters
        ----------
        is_deferring : `bool`
            Whether the request just deferring the interaction.
        is_ephemeral : `bool`
            Whether the request is ephemeral.
        """
        self = object.__new__(cls)
        self.interaction = interaction
        self.is_deferring = is_deferring
        self.is_ephemeral = is_ephemeral
        return self
    
    def __enter__(self):
        """Enters the context manager as deferring or responding if applicable."""
        interaction = self.interaction
        response_flag = interaction._response_flag
        
        if self.is_deferring:
            if not (response_flag & RESPONSE_FLAG_ACKNOWLEDGING_OR_ACKNOWLEDGED):
                response_flag |= RESPONSE_FLAG_DEFERRING
        else:
            if (not response_flag & RESPONSE_FLAG_RESPONDING_OR_RESPONDED) and \
                    (not response_flag & RESPONSE_FLAG_DEFERRING_OR_DEFERRED):
                response_flag |= RESPONSE_FLAG_RESPONDING
        
        interaction._response_flag = response_flag
        
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exits the context manager, marking the interaction as deferred or responded if no exception occurred."""
        interaction = self.interaction
        response_flag = interaction._response_flag
        if exc_type is None:
            if self.is_ephemeral:
                if not response_flag & RESPONSE_FLAG_ACKNOWLEDGED:
                    response_flag ^= RESPONSE_FLAG_EPHEMERAL
            
            if self.is_deferring:
                if response_flag & RESPONSE_FLAG_DEFERRING:
                    response_flag ^= RESPONSE_FLAG_DEFERRING
                    response_flag |= RESPONSE_FLAG_DEFERRED
            else:
                if response_flag & RESPONSE_FLAG_RESPONDING:
                    response_flag ^= RESPONSE_FLAG_RESPONDING
                    response_flag |= RESPONSE_FLAG_RESPONDED
        
        else:
            if self.is_deferring:
                if response_flag & RESPONSE_FLAG_DEFERRING:
                    response_flag ^= RESPONSE_FLAG_DEFERRING
            else:
                if response_flag & RESPONSE_FLAG_RESPONDING:
                    response_flag ^= RESPONSE_FLAG_RESPONDING
        
        interaction._response_flag = response_flag
        return False



__all__ = ('EmbeddedActivityState',)

from ..activity import ActivityRich
from ..core import CHANNELS, EMBEDDED_ACTIVITY_STATES, GUILDS
from ..user import create_partial_user_from_id


EMBEDDED_ACTIVITY_UPDATE_NONE = 0
EMBEDDED_ACTIVITY_UPDATE_CREATE = 1
EMBEDDED_ACTIVITY_UPDATE_DELETE = 2
EMBEDDED_ACTIVITY_UPDATE_UPDATE = 3
EMBEDDED_ACTIVITY_UPDATE_USER_ADD = 4
EMBEDDED_ACTIVITY_UPDATE_USER_DELETE = 5


def _add_embedded_activity_state_to_guild_cache(embedded_activity_state):
    """
    Adds the embedded activity to it's guild.
    
    Parameters
    ----------
    embedded_activity_state : ``EmbeddedActivityState``
        The embedded activity to add to it's guild's cache.
    """
    try:
        guild = GUILDS[embedded_activity_state.guild_id]
    except KeyError:
        pass
    else:
        embedded_activity_states = guild._embedded_activity_states
        if (embedded_activity_states is None):
            embedded_activity_states = set()
            guild._embedded_activity_states = embedded_activity_states
        
        embedded_activity_states.add(embedded_activity_state)


def _remove_embedded_activity_state_from_guild_cache(embedded_activity_state):
    """
    Adds the embedded activity to it's guild.
    
    Parameters
    ----------
    embedded_activity_state : ``EmbeddedActivityState``
        The embedded activity to add to it's guild's cache.
    """
    try:
        guild = GUILDS[embedded_activity_state.guild_id]
    except KeyError:
        pass
    else:
        embedded_activity_states = guild._embedded_activity_states
        if (embedded_activity_states is not None):
            try:
                embedded_activity_states.remove(embedded_activity_state)
            except KeyError:
                pass
            else:
                if not embedded_activity_state:
                    guild._embedded_activity_states = None


def difference_handle_embedded_activity_update_event(data):
    """
    Handles embedded activity events returning the embedded activity state and the changes with it.
    
    Parameters
    ----------
    data : `dict` of (`str`, `Any`) items
        Embedded activity update event.
    
    Returns
    -------
    embedded_activity_state : ``EmbeddedActivityState``
        The updated embedded activity state.
    
    changes : `list` of `tuple` (`int`, `Any`)
        Change entries. Each tuple` has `2` elements:
        
        - The action identifier of the change:
        
            Can be one of the following:
            
            +---------------------------------------+-------+
            | Respective name                       | Value |
            +=======================================+=======+
            | EMBEDDED_ACTIVITY_UPDATE_NONE         | 0     |
            +---------------------------------------+-------+
            | EMBEDDED_ACTIVITY_UPDATE_CREATE       | 1     |
            +---------------------------------------+-------+
            | EMBEDDED_ACTIVITY_UPDATE_DELETE       | 2     |
            +---------------------------------------+-------+
            | EMBEDDED_ACTIVITY_UPDATE_UPDATE       | 3     |
            +---------------------------------------+-------+
            | EMBEDDED_ACTIVITY_UPDATE_USER_ADD     | 4     |
            +---------------------------------------+-------+
            | EMBEDDED_ACTIVITY_UPDATE_USER_DELETE  | 5     |
            +---------------------------------------+-------+
        
        - The value in the context of the action.
            
            If `action` is `EMBEDDED_ACTIVITY_UPDATE_UPDATE`, this value will contain the updated attributes of the
            activity.
            
            +-------------------+-----------------------------------+
            | Keys              | Values                            |
            +===================+===================================+
            | application_id    | `int`                             |
            +-------------------+-----------------------------------+
            | assets            | `None`, ``ActivityAssets``        |
            +-------------------+-----------------------------------+
            | created_at        | `datetime`                        |
            +-------------------+-----------------------------------+
            | details           | `None`, `str`                     |
            +-------------------+-----------------------------------+
            | flags             | ``ActivityFlag``                  |
            +-------------------+-----------------------------------+
            | name              | `str`                             |
            +-------------------+-----------------------------------+
            | party             | `None`, ``ActivityParty``         |
            +-------------------+-----------------------------------+
            | secrets           | `None`, ``ActivitySecrets``       |
            +-------------------+-----------------------------------+
            | session_id        | `None`, `str`                     |
            +-------------------+-----------------------------------+
            | state             | `None`, `str`                     |
            +-------------------+-----------------------------------+
            | sync_id           | `None`, `str`                     |
            +-------------------+-----------------------------------+
            | timestamps        | `None`, `ActivityTimestamps``     |
            +-------------------+-----------------------------------+
            | url               | `None`, `str`                     |
            +-------------------+-----------------------------------+
            
            If `action` is `EMBEDDED_ACTIVITY_UPDATE_USER_ADD`, `EMBEDDED_ACTIVITY_UPDATE_USER_DELETE`, it will
            contain the joined or left user's identifier.
    """
    embedded_activity_state, is_created = EmbeddedActivityState(data, None)
    
    changes = []
    
    # If the object was just created, we have 2 cases:
    # Just came to cache as a deleted activity / it was created
    if is_created:
        if embedded_activity_state.users:
            _add_embedded_activity_state_to_guild_cache(embedded_activity_state)
            
            changes.append((EMBEDDED_ACTIVITY_UPDATE_CREATE, None))
        else:
            changes.append((EMBEDDED_ACTIVITY_UPDATE_DELETE, None))
    
    else:
        joined_user_ids, left_user_ids = embedded_activity_state._difference_update_use_ids(data)
        for user_id in joined_user_ids:
            changes.append((EMBEDDED_ACTIVITY_UPDATE_USER_ADD, user_id))
        
        for user_id in left_user_ids:
            changes.append((EMBEDDED_ACTIVITY_UPDATE_USER_DELETE, user_id))
        
        if embedded_activity_state.user_ids:
            activity_old_attributes = embedded_activity_state.activity._difference_update_attributes(
                data['embedded_activity'],
            )
            if activity_old_attributes:
                changes.append((EMBEDDED_ACTIVITY_UPDATE_UPDATE, activity_old_attributes))
        else:
            embedded_activity_state.activity._update_attributes(data['embedded_activity'])
            
            _remove_embedded_activity_state_from_guild_cache(embedded_activity_state)
            changes.append((EMBEDDED_ACTIVITY_UPDATE_DELETE, None))
    
    return embedded_activity_state, changes


def handle_embedded_update_event(data):
    """
    Handles an embedded activity event an iterates over the resulted events.
    
    Parameters
    ----------
    data : `dict` of (`str`, `Any`) items
        Embedded activity update event.
    """
    embedded_activity_state, is_created = EmbeddedActivityState(data, None)
    if is_created:
        _add_embedded_activity_state_to_guild_cache(embedded_activity_state)
    else:
        embedded_activity_state._update_user_ids(data)
        embedded_activity_state.activity._update_attributes(data['embedded_activity'])
        
        if not embedded_activity_state.user_ids:
            _remove_embedded_activity_state_from_guild_cache(embedded_activity_state)


class EmbeddedActivityState:
    """
    Attributes
    ----------
    activity : ``ActivityRich``
        The embedded activity.
    channel_id : `int`
        The respective channel's identifier.
    guild_id : `int`
        The respective guild's identifier.
    user_ids : `set` of `int`
        The joined users' identifiers.
    
    Notes
    -----
    Embedded activity instances are weakreferable.
    """
    __slots__ = ('__weakref__', 'activity', 'channel_id', 'guild_id', 'user_ids')
    
    
    def __new__(cls, data, guild_id):
        """
        Creates a new embedded activity state instance.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received data.
        guild_id : `None`, `int`
            The guild's identifier where the activity is.
        
        Returns
        -------
        self : ``EmbeddedActivityState``
            The instance itself.
        is_created : `bool`
            Whether the instance was just created.
        """
        if guild_id is None:
            guild_id = data.get('guild_id', None)
            if guild_id is None:
                guild_id = 0
            else:
                guild_id = int(guild_id)
        
        channel_id = int(data['channel_id'])
        activity_data = data['embedded_activity']
        application_id = activity_data.get('application_id', None)
        if application_id is None:
            application_id = 0
        else:
            application_id = int(application_id)
        
        key = (guild_id, channel_id, application_id)
        
        try:
            self = EMBEDDED_ACTIVITY_STATES[key]
        except KeyError:
            self = object.__new__(cls)
            self.channel_id = channel_id
            self.guild_id = guild_id
            self.activity = ActivityRich.from_data(activity_data)
            self.user_ids = set(int(user_id) for user_id in data['users'])
            
            EMBEDDED_ACTIVITY_STATES[key] = self
            is_created = True
        
        else:
            is_created = False
        
        return self, is_created
    
    
    def __repr__(self):
        """Returns the embedded activity's representation."""
        repr_parts = [
            '<', self.__class__.__name__,
            ' guild_id=', repr(self.guild_id),
            ', channel_id=', repr(self.channel_id),
            ', user_ids=', repr(self.user_ids),
            '>',
        ]
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two embedded activities are the same."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.activity.application_id != other.activity.application_id:
            return False
        
        if self.channel_id != other.channel_id:
            return False
        
        if self.guild_id != other.guild_id:
            return False
        
        
    def __hash__(self):
        """Returns the hash value of the embedded activity state."""
        return self.guild_id ^ self.channel_id ^ self.activity.application_id
    
    
    @property
    def guild(self):
        """
        Returns the guild where the embedded activity is. The guild must be cached.
        
        Returns
        -------
        guild : `None`, ``Guild``
        """
        return GUILDS.get(self.guild_id, None)
    
    
    @property
    def channel(self):
        """
        Returns the channel where the embedded activity is. The channel must be cached.
        
        Returns
        -------
        channel : `None`, ``ChannelBase``
        """
        return CHANNELS.get(self.channel_id, None)
    
    
    @property
    def users(self):
        """
        Returns the users of the embedded activity.
        
        Returns
        -------
        users : `list` of ``ClientUserBase``
        """
        return [create_partial_user_from_id(user_id) for user_id in self.user_ids]
    
    
    def _update_user_ids(self, data):
        """
        Updates the user id-s of the embedded activity state.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Embedded activity state data.
        """
        self.user_ids = set(int(user_id) for user_id in data['users'])
    
    
    def _difference_update_use_ids(self, data):
        """
        Difference updates the user id-s of the embedded activity state.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Embedded activity state data.
        
        Returns
        -------
        joined_user_ids : `set` of `int`
            The joined users' identifiers.
        left_user_ids : `set` of `int`
            The left users' identifiers.
        """
        new_user_ids = set(int(user_id) for user_id in data['users'])
        old_user_ids = self.user_ids
        self.user_ids = new_user_ids
        
        return new_user_ids - old_user_ids, old_user_ids - new_user_ids



__all__ = ('ComponentCommand', )

from scarletio import copy_docs

from ...discord.events.handling_helpers import Router, check_name, route_name, route_value

from .converters import get_component_command_parameter_converters
from .custom_id_based_command import (
    CustomIdBasedCommand, _validate_custom_ids, _validate_name, split_and_check_satisfaction
)
from .exceptions import handle_command_exception
from .responding import process_command_coroutine
from .utils import _check_maybe_route
from .wrappers import SlasherCommandWrapper


try:
    # CPython
    from re import Pattern
except ImportError:
    # ChadPython (PyPy)
    from re import _pattern_type as Pattern

COMMAND_TARGETS_COMPONENT_COMMAND = frozenset((
    'component',
))

class ComponentCommand(CustomIdBasedCommand):
    """
    A command, which is called if a message component interaction is received with a matched `custom_id`.
    
    Attributes
    ----------
    _command_function : `async-callable˛
        The command's function to call.
    _exception_handlers : `None`, `list` of `CoroutineFunction`
        Exception handlers added with ``.error`` to the interaction handler.
        
        Same as ``Slasher._exception_handlers``.
    
    _parent_reference : `None`, ``WeakReferer`` to ``SlasherApplicationCommand``
        The parent slasher of the component command.
    _parameter_converters : `tuple` of ``ParameterConverter``
        Parsers to parse command parameters.
    _string_custom_ids : `None`, `tuple` of `str`
        The custom id-s to wait for.
    _regex_custom_ids : `None`, `tuple` of `re.Pattern`.
        Regex pattern to match custom-ids.
    name : `str`
        The component commands name.
        
        Only used for debugging.
    
    Class Attributes
    ----------------
    COMMAND_COMMAND_NAME : `str`
        The command's name defining parameter's name.
    COMMAND_PARAMETER_NAMES : tuple of `str`
        All parameters names accepted by ``.__new__``
    COMMAND_NAME_NAME : `str`
        The command's command defining parameter's name.
    """
    __slots__ = ()
    
    
    def __new__(cls, func, custom_id, name=None, target=None):
        """
        Creates a new ``ComponentCommand`` with the given parameters
        
        Parameters
        ----------
        func : `None`, `async-callable`, Optional
            The function used as the command when using the respective slash command.
        custom_id : `str`, (`list`, `set`) of `str`, `tuple` of (`str`, (`list`, `set`) of `str`)
            Custom id to match by the component command.
        name : `str`, `None`, Optional
            The name of the component command.
        target : `str`, `None`, Optional
            The component command's target.
        
        Returns
        -------
        self : ``ComponentCommand``, ``Router``
        
        Raises
        ------
        TypeError
            - If `func` is not async callable, neither cannot be instanced to async.
            - If `func` accepts keyword only parameters.
            - If `func` accepts `*args`.
            - If `func` accepts `**kwargs`.
            - If `name` was not given neither as `None`, `str`.
            - If `custom_id`'s type is incorrect.
        ValueError:
            - If no `custom_id` was received.
            - If `custom_id` contains incorrect value.
            - If `target`'s value is not correct.
        """
        if (target is not None) and (target not in COMMAND_TARGETS_COMPONENT_COMMAND):
            raise ValueError(
                f'`target` can be `None` or any of `{COMMAND_TARGETS_COMPONENT_COMMAND!r}`\'s '
                f'values, got {target!r}'
            )
        
        if (func is not None) and isinstance(func, SlasherCommandWrapper):
            command, wrappers = func.fetch_function_and_wrappers_back()
        else:
            command = func
            wrappers = None
        
        # Check for routing.
        route_to = 0
        name, route_to = _check_maybe_route('name', name, route_to, _validate_name)
        custom_id, route_to = _check_maybe_route('custom_id', custom_id, route_to, _validate_custom_ids)
        
        command, parameter_converters = get_component_command_parameter_converters(command)
        
        if route_to:
            custom_id = route_value(custom_id, route_to)
            name = route_name(name, route_to)
            name = [check_name(command, sub_name) for sub_name in name]
            
            router = []
            
            for custom_id, name in zip(custom_id, name):
                string_custom_ids, regex_custom_ids = split_and_check_satisfaction(custom_id, parameter_converters)
                
                self = object.__new__(cls)
                self._command_function = command
                self._parameter_converters = parameter_converters
                self._string_custom_ids = string_custom_ids
                self._regex_custom_ids = regex_custom_ids
                self._parent_reference = None
                self._exception_handlers = None
                self.name = name
                
                if (wrappers is not None):
                    for wrapper in wrappers:
                        wrapper.apply(self)
                
                router.append(self)
            
            return Router(router)
        
        else:
            name = check_name(command, name)
            
            string_custom_ids, regex_custom_ids = split_and_check_satisfaction(custom_id, parameter_converters)
            
            self = object.__new__(cls)
            self._command_function = command
            self._parameter_converters = parameter_converters
            self._string_custom_ids = string_custom_ids
            self._regex_custom_ids = regex_custom_ids
            self._parent_reference = None
            self._exception_handlers = None
            self.name = name
            
            if (wrappers is not None):
                for wrapper in wrappers:
                    wrapper.apply(self)
            
            return self
    
    @copy_docs(CustomIdBasedCommand.__call__)
    async def __call__(self, client, interaction_event, regex_match):
        parameters = []
        
        for parameter_converter in self._parameter_converters:
            try:
                parameter = await parameter_converter(client, interaction_event, regex_match)
            except BaseException as err:
                exception = err
            else:
                parameters.append(parameter)
                continue
            
            await handle_command_exception(
                self,
                client,
                interaction_event,
                exception,
            )
            return
        
        command_coroutine = self._command_function(*parameters)
        
        try:
            await process_command_coroutine(client, interaction_event, False, command_coroutine)
        except BaseException as err:
            exception = err
        else:
            return
        
        await handle_command_exception(
            self,
            client,
            interaction_event,
            exception,
        )
        return
